import uuid
import typer
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from skyplane.api.config import TransferConfig
from skyplane.api.dataplane import Dataplane
from skyplane.api.provisioner import Provisioner
from skyplane.api.obj_store import ObjectStore
from skyplane.api.usage import get_clientid
from skyplane.obj_store.object_store_interface import ObjectStoreInterface
from skyplane.planner.planner import MulticastDirectPlanner
from skyplane.utils import logger
from skyplane.utils.definitions import tmp_log_dir
from skyplane.utils.path import parse_path

from skyplane.api.pipeline import Pipeline

if TYPE_CHECKING:
    from skyplane.api.config import AWSConfig, AzureConfig, GCPConfig, TransferConfig, IBMCloudConfig


class SkyplaneClient:
    """Client for initializing cloud provider configurations."""

    def __init__(
        self,
        aws_config: Optional["AWSConfig"] = None,
        azure_config: Optional["AzureConfig"] = None,
        gcp_config: Optional["GCPConfig"] = None,
        ibmcloud_config: Optional["IBMCloudConfig"] = None,
        transfer_config: Optional[TransferConfig] = None,
        log_dir: Optional[str] = None,
    ):
        """
        :param aws_config: aws cloud configurations
        :type aws_config: class AWSConfig (optional)
        :param azure_config: azure cloud configurations
        :type azure_config: class AzureConfig (optional)
        :param gcp_config: gcp cloud configurations
        :type gcp_config: class GCPConfig (optional)
        :param ibmcloud_config: ibm cloud configurations
        :type ibmcloud_config: class IBMCloudConfig (optional)
        :param transfer_config: transfer configurations
        :type transfer_config: class TransferConfig (optional)
        :param log_dir: path to store transfer logs
        :type log_dir: str (optional)
        """
        self.clientid = get_clientid()
        self.aws_auth = aws_config.make_auth_provider() if aws_config else None
        self.azure_auth = azure_config.make_auth_provider() if azure_config else None
        self.gcp_auth = gcp_config.make_auth_provider() if gcp_config else None
        self.ibmcloud_auth = ibmcloud_config.make_auth_provider() if ibmcloud_config else None
        self.transfer_config = transfer_config if transfer_config else TransferConfig()
        self.log_dir = (
            tmp_log_dir / "transfer_logs" / f"{datetime.now().strftime('%Y%m%d_%H%M%S')}-{uuid.uuid4().hex[:8]}"
            if log_dir is None
            else Path(log_dir)
        )

        # set up logging
        self.log_dir.mkdir(parents=True, exist_ok=True)
        logger.open_log_file(self.log_dir / "client.log")
        typer.secho(f"Logging to: {self.log_dir / 'client.log'}", fg="bright_black")

        self.provisioner = Provisioner(
            host_uuid=self.clientid,
            aws_auth=self.aws_auth,
            azure_auth=self.azure_auth,
            gcp_auth=self.gcp_auth,
            ibmcloud_auth=self.ibmcloud_auth,
        )

    def pipeline(self, planning_algorithm: Optional[str] = "direct", max_instances: Optional[int] = 1, debug=False):
        """Create a pipeline object to queue jobs"""
        return Pipeline(
            planning_algorithm=planning_algorithm,
            max_instances=max_instances,
            clientid=self.clientid,
            provisioner=self.provisioner,
            transfer_config=self.transfer_config,
            debug=debug,
        )

    def copy(self, src: str, dst: str, recursive: bool = False):
        """
        A simple version of Skyplane copy. It automatically waits for transfer to complete
        (the main thread is blocked) and deprovisions VMs at the end.

        :param src: Source prefix to copy from
        :type src: str
        :param dst: The destination of the transfer
        :type dst: str
        :param recursive: If true, will copy objects at folder prefix recursively (default: False)
        :type recursive: bool
        :param num_vms: The maximum number of instances to use per region (default: 1)
        :type num_vms: int
        """

        pipeline = self.pipeline()
        pipeline.queue_copy(src, dst, recursive=recursive)
        pipeline.start(progress=True)

    def object_store(self):
        return ObjectStore()



import subprocess
from datetime import datetime
from multiprocessing import Queue
from os import PathLike
from pathlib import Path
from typing import Dict, Optional

from skyplane.utils import logger

from skyplane.chunk import ChunkRequest, ChunkState
from skyplane.gateway.gateway_queue import GatewayQueue


class ChunkStore:
    def __init__(self, chunk_dir: PathLike):
        self.chunk_dir = Path(chunk_dir)
        self.chunk_dir.mkdir(parents=True, exist_ok=True)

        self.region_key_upload_id_mappings: Dict[str, str] = {}

        # delete existing chunks
        for chunk_file in self.chunk_dir.glob("*.chunk"):
            logger.warning(f"Deleting existing chunk file {chunk_file}")
            chunk_file.unlink()

        # queues of incoming chunk requests for each partition from gateway API (passed to operator graph)
        self.chunk_requests: Dict[str, GatewayQueue] = {}

        # queue of chunk status updates coming from operators (passed to gateway API)
        self.chunk_status_queue: Queue[Dict] = Queue()

    def set_upload_ids_map(self, maps: Dict[str, str]):
        self.region_key_upload_id_mappings.update(maps)

    def get_upload_ids_map(self):
        return self.region_key_upload_id_mappings

    def add_partition(self, partition_id: str):
        """Create a queue for this partition."""
        if partition_id in self.chunk_requests:
            raise ValueError(f"Partition {partition_id} already exists")
        self.chunk_requests[partition_id] = GatewayQueue()

    def add_partition(self, partition_id: str, queue: GatewayQueue):
        """Create a queue for this partition."""
        print("Adding partition", partition_id, queue)
        if partition_id in self.chunk_requests:
            raise ValueError(f"Partition {partition_id} already exists")
        self.chunk_requests[partition_id] = queue
        print(self.chunk_requests)

    def add_chunk_request(self, chunk_request: ChunkRequest, state: ChunkState = ChunkState.registered):
        """Enqueue new chunk request from Gateway API
        :param chunk_request: ChunkRequest object
        :param state: ChunkState enum (registered, in_progress, complete)

        :return: size of Gateway queue
        """
        if chunk_request.chunk.partition_id not in self.chunk_requests:
            raise ValueError(
                f"Partition {chunk_request.chunk.partition_id} does not exist in {self.chunk_requests} - was the gateway program loaded?"
            )
        try:
            self.chunk_requests[chunk_request.chunk.partition_id].put_nowait(chunk_request)
        except Exception as e:
            print("Error adding chunk", e)
            return self.chunk_requests[chunk_request.chunk.partition_id].size(), False

        self.log_chunk_state(chunk_request, state)
        return self.chunk_requests[chunk_request.chunk.partition_id].size(), True

    def log_chunk_state(
        self,
        chunk_req: ChunkRequest,
        new_status: ChunkState,
        worker_id: Optional[int] = None,
        operator_handle: Optional[str] = None,
        metadata: Optional[Dict] = None,
    ):
        """Add dict containing chunk status change (coming from operators) to queue"""
        rec = {
            "chunk_id": chunk_req.chunk.chunk_id,
            "partition": chunk_req.chunk.partition_id,
            "state": new_status.name,
            "time": str(datetime.utcnow().isoformat()),
            "handle": operator_handle,
            "worker_id": worker_id,
        }
        if metadata is not None:
            rec.update(metadata)
        self.chunk_status_queue.put(rec)

    # Memory space calculation
    def remaining_bytes(self):
        try:
            remaining_bytes = (
                int(subprocess.check_output(["df", "-k", "--output=avail", self.chunk_dir]).decode().strip().split()[-1]) * 1024
            )
            return remaining_bytes
        except Exception as e:
            raw_output = subprocess.check_output(["df", "-k", "--output=avail", self.chunk_dir]).decode().strip()
            print(f"{str(e)}: failed to parse {raw_output}")
            return 0

    def get_upload_id_map_path(self) -> Path:
        return self.chunk_dir / f"upload_id_map.json"

    def get_chunk_file_path(self, chunk_id: str) -> Path:
        return self.chunk_dir / f"{chunk_id}.chunk"



#
# (C) Copyright Cloudlab URV 2020
# (C) Copyright IBM Corp. 2023
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import uuid


DEFAULT_CONFIG_KEYS = {
    "master_profile_name": "bx2-2x8",
    "worker_profile_name": "bx2-2x8",
    "boot_volume_profile": "general-purpose",
    "ssh_username": "root",
    "ssh_password": str(uuid.uuid4()),
    "ssh_key_filename": "~/.ssh/id_rsa",
    "delete_on_dismantle": True,
    "max_workers": 100,
    "worker_processes": 2,
    "boot_volume_capacity": 100,
}

VPC_ENDPOINT = "https://{}.iaas.cloud.ibm.com"

REGIONS = ["us-south", "us-east", "eu-de", "jp-tok", "jp-osa", "au-syd", "eu-gb", "br-sao", "ca-tor"]


def load_config(config_data):
    if "ibm" in config_data and config_data["ibm"] is not None:
        config_data["ibm_gen2"].update(config_data["ibm"])

    for key in DEFAULT_CONFIG_KEYS:
        if key not in config_data["ibm_gen2"]:
            config_data["ibm_gen2"][key] = DEFAULT_CONFIG_KEYS[key]

    if "profile_name" in config_data["ibm_gen2"]:
        config_data["worker_profile_name"] = config_data["profile_name"]

    if "region" not in config_data["ibm_gen2"] and "endpoint" not in config_data["ibm_gen2"]:
        msg = "'region' or 'endpoint' parameter is mandatory in 'ibm_gen2' section of the configuration"
        raise Exception(msg)

    if "region" in config_data["ibm_gen2"]:
        region = config_data["ibm_gen2"]["region"]
        if region not in REGIONS:
            msg = f"'region' conig parameter in 'ibm_gen2' section must be one of {REGIONS}"
            raise Exception(msg)
        config_data["ibm_gen2"]["endpoint"] = VPC_ENDPOINT.format(region)

    config_data["ibm_gen2"]["endpoint"] = config_data["ibm_gen2"]["endpoint"].replace("/v1", "")



import json
import time
import os
import threading
from collections import defaultdict, Counter
from datetime import datetime
from functools import partial
from datetime import datetime

import nacl.secret
import nacl.utils
import urllib3
from typing import TYPE_CHECKING, Dict, List, Optional

from skyplane import compute
from skyplane.api.tracker import TransferProgressTracker, TransferHook
from skyplane.api.transfer_job import CopyJob, SyncJob, TransferJob
from skyplane.api.config import TransferConfig

from skyplane.planner.planner import MulticastDirectPlanner, DirectPlannerSourceOneSided, DirectPlannerDestOneSided
from skyplane.planner.topology import TopologyPlanGateway
from skyplane.utils import logger
from skyplane.utils.definitions import gateway_docker_image, tmp_log_dir
from skyplane.utils.fn import PathLike, do_parallel

from skyplane.api.dataplane import Dataplane

if TYPE_CHECKING:
    from skyplane.api.provisioner import Provisioner


class Pipeline:
    """A pipeline object stores and executes a set of transfer jobs."""

    def __init__(
        self,
        clientid: str,
        provisioner: "Provisioner",
        transfer_config: TransferConfig,
        # cloud_regions: dict,
        max_instances: Optional[int] = 1,
        n_connections: Optional[int] = 64,
        planning_algorithm: Optional[str] = "direct",
        debug: Optional[bool] = False,
    ):
        """
        :param clientid: the uuid of the local host to create the dataplane
        :type clientid: str
        :param provisioner: the provisioner to launch the VMs
        :type provisioner: Provisioner
        :param transfer_config: the configuration during the transfer
        :type transfer_config: TransferConfig
        """
        self.clientid = clientid
        # self.cloud_regions = cloud_regions
        # TODO: set max instances with VM CPU limits and/or config
        self.max_instances = max_instances
        self.n_connections = n_connections
        self.provisioner = provisioner
        self.transfer_config = transfer_config
        self.http_pool = urllib3.PoolManager(retries=urllib3.Retry(total=3))
        self.provisioning_lock = threading.Lock()
        self.provisioned = False
        self.transfer_dir = tmp_log_dir / "transfer_logs" / datetime.now().strftime("%Y%m%d_%H%M%S")
        self.transfer_dir.mkdir(exist_ok=True, parents=True)

        # dataplane
        self.dataplane = None

        # planner
        self.planning_algorithm = planning_algorithm
        if self.planning_algorithm == "direct":
            self.planner = MulticastDirectPlanner(self.max_instances, self.n_connections, self.transfer_config)
        elif self.planning_algorithm == "src_one_sided":
            self.planner = DirectPlannerSourceOneSided(self.max_instances, self.n_connections, self.transfer_config)
        elif self.planning_algorithm == "dst_one_sided":
            self.planner = DirectPlannerDestOneSided(self.max_instances, self.n_connections, self.transfer_config)
        else:
            raise ValueError(f"No such planning algorithm {planning_algorithm}")

        # transfer logs
        self.transfer_dir = tmp_log_dir / "transfer_logs" / datetime.now().strftime("%Y%m%d_%H%M%S")
        self.transfer_dir.mkdir(exist_ok=True, parents=True)
        self.debug = debug

        # pending tracker tasks
        self.jobs_to_dispatch: List[TransferJob] = []
        self.pending_transfers: List[TransferProgressTracker] = []
        self.bound_nodes: Dict[TopologyPlanGateway, compute.Server] = {}

    def create_dataplane(self, debug):
        # create plan from set of jobs scheduled
        topo = self.planner.plan(self.jobs_to_dispatch)

        # create dataplane from plan
        dp = Dataplane(self.clientid, topo, self.provisioner, self.transfer_config, str(self.transfer_dir), debug=debug)
        return dp

    def start(self, debug=False, progress=False):
        ## create plan from set of jobs scheduled
        # topo = self.planner.plan(self.jobs_to_dispatch)

        ## create dataplane from plan
        # dp = Dataplane(self.clientid, topo, self.provisioner, self.transfer_config, self.transfer_dir, debug=debug)
        dp = self.create_dataplane(debug)
        try:
            dp.provision(spinner=True)
            if progress:
                from skyplane.cli.impl.progress_bar import ProgressBarTransferHook

                tracker = dp.run_async(self.jobs_to_dispatch, hooks=ProgressBarTransferHook(dp.topology.dest_region_tags))
            else:
                tracker = dp.run_async(self.jobs_to_dispatch)

            # wait for job to finish
            tracker.join()

            # copy gateway logs
            if debug:
                dp.copy_gateway_logs()
        except Exception as e:
            dp.copy_gateway_logs()
        dp.deprovision(spinner=True)
        return dp

    def start_async(self, debug=False):
        dp = self.create_dataplane(debug)
        try:
            dp.provision(spinner=False)
            tracker = dp.run_async(self.jobs_to_dispatch)
            if debug:
                dp.copy_gateway_logs()
            return tracker
        except Exception as e:
            dp.copy_gateway_logs()
            return

    def queue_copy(
        self,
        src: str,
        dst: str or List[str],
        recursive: bool = False,
    ) -> str:
        """
        Add a copy job to job list.
        Return the uuid of the job.

        :param src: source prefix to copy from
        :type src: str
        :param dst: the destination of the transfer
        :type dst: str
        :param recursive: if true, will copy objects at folder prefix recursively (default: False)
        :type recursive: bool
        """
        if isinstance(dst, str):
            dst = [dst]
        job = CopyJob(src, dst, recursive, requester_pays=self.transfer_config.requester_pays)
        logger.fs.debug(f"[SkyplaneClient] Queued copy job {job}")
        self.jobs_to_dispatch.append(job)
        return job.uuid

    def queue_sync(
        self,
        src: str,
        dst: str or List[str],
    ) -> str:
        """
        Add a sync job to job list.
        Return the uuid of the job.

        :param src: Source prefix to copy from
        :type src: str
        :param dst: The destination of the transfer
        :type dst: str
        :param recursive: If true, will copy objects at folder prefix recursively (default: False)
        :type recursive: bool
        """
        if isinstance(dst, str):
            dst = [dst]
        job = SyncJob(src, dst, requester_pays=self.transfer_config.requester_pays)
        logger.fs.debug(f"[SkyplaneClient] Queued sync job {job}")
        self.jobs_to_dispatch.append(job)
        return job.uuid

    def estimate_total_cost(self):
        """Estimate total cost of queued jobs"""
        total_size = 0
        for job in self.jobs_to_dispatch:
            total_size += job.size_gb()

        # get planner topology
        topo = self.planner.plan(self.jobs_to_dispatch)

        # return size
        return total_size * topo.cost_per_gb



from typing import Optional, List, Tuple
import json
from collections import defaultdict


class GatewayOperator:
    def __init__(self, op_type):
        self.op_type = op_type
        self.children = []
        self.handle = None

    def add_children(self, children):
        self.children.extend(children)

    def add_child(self, child):
        self.children.append(child)

    def set_handle(self, handle: str):
        self.handle = handle

    def to_dict(self):
        if len(self.children) == 0:
            return {**self.__dict__, **{"children": []}}

        return {**self.__dict__, **{"children": [child.to_dict() for child in self.children]}}

    def to_json(self):
        return json.dumps(self.to_dict())

    def __repr__(self):
        return self.to_json()


class GatewaySend(GatewayOperator):
    def __init__(
        self,
        target_gateway_id: str,
        region: str,
        num_connections: int = 32,
        compress: bool = False,
        encrypt: bool = False,
        private_ip: bool = False,
    ):
        super().__init__("send")
        self.target_gateway_id = target_gateway_id  # gateway to send to
        self.region = region  # region to send to
        self.num_connections = num_connections  # default this for now
        self.compress = compress
        self.encrypt = encrypt
        self.private_ip = private_ip  # whether to send to private or public IP (private for GCP->GCP)


class GatewayReceive(GatewayOperator):
    def __init__(self, decompress: bool = False, decrypt: bool = False, max_pending_chunks: int = 1000):
        super().__init__("receive")
        self.decompress = decompress
        self.decrypt = decrypt
        self.max_pending_chunks = max_pending_chunks


class GatewayGenData(GatewayOperator):
    def __init__(self, size_mb: int):
        super().__init__("gen_data")
        self.size_mb = size_mb


class GatewayReadObjectStore(GatewayOperator):
    def __init__(self, bucket_name: str, bucket_region: str, num_connections: int = 32):
        super().__init__("read_object_store")
        self.bucket_name = bucket_name
        self.bucket_region = bucket_region
        self.num_connections = num_connections


class GatewayWriteObjectStore(GatewayOperator):
    def __init__(self, bucket_name: str, bucket_region: str, num_connections: int = 32, key_prefix: Optional[str] = ""):
        super().__init__("write_object_store")
        self.bucket_name = bucket_name
        self.bucket_region = bucket_region
        self.num_connections = num_connections
        self.key_prefix = key_prefix


class GatewayWriteLocal(GatewayOperator):
    def __init__(self, path: Optional[str] = None):
        super().__init__("write_local")
        self.path = path


class GatewayMuxAnd(GatewayOperator):
    def __init__(self):
        super().__init__("mux_and")


class GatewayMuxOr(GatewayOperator):
    def __init__(self):
        super().__init__("mux_or")


class GatewayProgram:

    """
    GatewayProgram is a class that defines the local topology for a gateway (i.e. how to handle chunks)
    We intend to extend GatewayProgram to eventually support operations on chunk data.
    """

    ip_address = None

    def __init__(self):
        self._plan = defaultdict(list)
        self._ops = {}

    def get_operators(self) -> List[GatewayOperator]:
        return list(self._ops.values())

    def add_operators(self, ops: List[GatewayOperator], parent_handle: Optional[str] = None, partition_id: Optional[str] = "default"):
        parent_op = self._ops[parent_handle] if parent_handle else None
        ops_handles = []
        for op in ops:
            ops_handles.append(self.add_operator(op, parent_op, partition_id))

        return ops_handles

    def add_operator(self, op: GatewayOperator, parent_handle: Optional[str] = None, partition_id: Optional[str] = "default"):
        parent_op = self._ops[parent_handle] if parent_handle else None
        if not parent_op:  # root operation
            self._plan[partition_id].append(op)
        else:
            parent_op.add_child(op)
        op.set_handle(f"operator_{len(list(self._ops.keys()))}")
        self._ops[op.handle] = op
        return op.handle

    def to_dict(self):
        """
        Return dictionary representation of all partitions and gateway partitions.
        Partitions with equivalent programs are grouped together to save space.
        """
        program_all = []
        for partition_id, op_list in self._plan.items():
            # build gateway program representation
            program = []
            for op in op_list:
                program.append(op.to_dict())

            # check if any existing
            exists = False
            for p in program_all:
                if p["value"] == program:  # equivalent partition exists
                    p["partitions"].append(partition_id)
                    exists = True
                    break
            if not exists:
                program_all.append({"value": program, "partitions": [partition_id]})

        return program_all

    def to_json(self):
        return json.dumps(self.to_dict())



from pathlib import Path

from skyplane import compute, exceptions
from skyplane.api.client import SkyplaneClient
from skyplane.api.config import TransferConfig, AWSConfig, AzureConfig, GCPConfig
from skyplane.api.dataplane import Dataplane
from skyplane.api.pipeline import Pipeline
from skyplane.api.tracker import TransferHook

__version__ = "0.3.2"
__root__ = Path(__file__).parent.parent
__all__ = [
    "__version__",
    "__root__",
    # modules
    "compute",
    "exceptions",
    # API
    "SkyplaneClient",
    "Dataplane",
    "Pipeline",
    "TransferConfig",
    "AWSConfig",
    "AzureConfig",
    "GCPConfig",
    "TransferHook",
]



from collections import defaultdict
from importlib.resources import path
from typing import Dict, List, Optional, Tuple, Tuple
import os
import csv

from skyplane import compute
from skyplane.api.config import TransferConfig
from skyplane.utils import logger

from skyplane.planner.topology import TopologyPlan
from skyplane.gateway.gateway_program import (
    GatewayProgram,
    GatewayMuxOr,
    GatewayMuxAnd,
    GatewayReadObjectStore,
    GatewayWriteObjectStore,
    GatewayReceive,
    GatewaySend,
)

from skyplane.api.transfer_job import TransferJob
import json

from skyplane.utils.fn import do_parallel
from skyplane.config_paths import config_path, azure_standardDv5_quota_path, aws_quota_path, gcp_quota_path
from skyplane.config import SkyplaneConfig


class Planner:
    def __init__(self, transfer_config: TransferConfig, quota_limits_file: Optional[str] = None):
        self.transfer_config = transfer_config
        self.config = SkyplaneConfig.load_config(config_path)
        self.n_instances = self.config.get_flag("max_instances")

        # Loading the quota information, add ibm cloud when it is supported
        quota_limits = {}
        if quota_limits_file is not None:
            with open(quota_limits_file, "r") as f:
                quota_limits = json.load(f)
        else:
            if os.path.exists(aws_quota_path):
                with aws_quota_path.open("r") as f:
                    quota_limits["aws"] = json.load(f)
            if os.path.exists(azure_standardDv5_quota_path):
                with azure_standardDv5_quota_path.open("r") as f:
                    quota_limits["azure"] = json.load(f)
            if os.path.exists(gcp_quota_path):
                with gcp_quota_path.open("r") as f:
                    quota_limits["gcp"] = json.load(f)
        self.quota_limits = quota_limits

        # Loading the vcpu information - a dictionary of dictionaries
        # {"cloud_provider": {"instance_name": vcpu_cost}}
        self.vcpu_info = defaultdict(dict)
        with path("skyplane.data", "vcpu_info.csv") as file_path:
            with open(file_path, "r") as csvfile:
                reader = csv.reader(csvfile)
                next(reader)  # Skip the header row

                for row in reader:
                    instance_name, cloud_provider, vcpu_cost = row
                    vcpu_cost = int(vcpu_cost)
                    self.vcpu_info[cloud_provider][instance_name] = vcpu_cost

    def plan(self) -> TopologyPlan:
        raise NotImplementedError

    def _vm_to_vcpus(self, cloud_provider: str, vm: str) -> int:
        """Gets the vcpu_cost of the given vm instance (instance_name)

        :param cloud_provider: name of the cloud_provider
        :type cloud_provider: str
        :param instance_name: name of the vm instance
        :type instance_name: str
        """
        return self.vcpu_info[cloud_provider][vm]

    def _get_quota_limits_for(self, cloud_provider: str, region: str, spot: bool = False) -> Optional[int]:
        """Gets the quota info from the saved files. Returns None if quota_info isn't loaded during `skyplane init`
        or if the quota info doesn't include the region.

        :param cloud_provider: name of the cloud provider of the region
        :type cloud_provider: str
        :param region: name of the region for which to get the quota for
        :type region: int
        :param spot: whether to use spot specified by the user config (default: False)
        :type spot: bool
        """
        quota_limits = self.quota_limits.get(cloud_provider, None)
        if not quota_limits:
            # User needs to reinitialize to save the quota information
            return None
        if cloud_provider == "gcp":
            region_family = "-".join(region.split("-")[:2])
            if region_family in quota_limits:
                return quota_limits[region_family]
        elif cloud_provider == "azure":
            if region in quota_limits:
                return quota_limits[region]
        elif cloud_provider == "aws":
            for quota in quota_limits:
                if quota["region_name"] == region:
                    return quota["spot_standard_vcpus"] if spot else quota["on_demand_standard_vcpus"]
        return None

    def _calculate_vm_types(self, region_tag: str) -> Optional[Tuple[str, int]]:
        """Calculates the largest allowed vm type according to the regional quota limit as well as
        how many of these vm types can we launch to avoid QUOTA_EXCEEDED errors. Returns None if quota
        information wasn't properly loaded or allowed vcpu list is wrong.

        :param region_tag: tag of the node we are calculating the above for, example -> "aws:us-east-1"
        :type region_tag: str
        """
        cloud_provider, region = region_tag.split(":")

        # Get the quota limit
        quota_limit = self._get_quota_limits_for(
            cloud_provider=cloud_provider, region=region, spot=getattr(self.transfer_config, f"{cloud_provider}_use_spot_instances")
        )

        config_vm_type = getattr(self.transfer_config, f"{cloud_provider}_instance_class")

        # No quota limits (quota limits weren't initialized properly during skyplane init)
        if quota_limit is None:
            logger.warning(
                f"Quota limit file not found for {region_tag}. Try running `skyplane init --reinit-{cloud_provider}` to load the quota information"
            )
            # return default instance type and number of instances
            return config_vm_type, self.n_instances

        config_vcpus = self._vm_to_vcpus(cloud_provider, config_vm_type)
        if config_vcpus <= quota_limit:
            return config_vm_type, quota_limit // config_vcpus

        vm_type, vcpus = None, None
        for instance_name, vcpu_cost in sorted(self.vcpu_info[cloud_provider].items(), key=lambda x: x[1], reverse=True):
            if vcpu_cost <= quota_limit:
                vm_type, vcpus = instance_name, vcpu_cost
                break

        # shouldn't happen, but just in case we use more complicated vm types in the future
        assert vm_type is not None and vcpus is not None

        # number of instances allowed by the quota with the selected vm type
        n_instances = quota_limit // vcpus
        logger.warning(
            f"Falling back to instance class `{vm_type}` at {region_tag} "
            f"due to cloud vCPU limit of {quota_limit}. You can visit https://skyplane.org/en/latest/increase_vcpus.html "
            "to learn more about how to increase your cloud vCPU limits for any cloud provider."
        )
        return (vm_type, n_instances)

    def _get_vm_type_and_instances(
        self, src_region_tag: Optional[str] = None, dst_region_tags: Optional[List[str]] = None
    ) -> Tuple[Dict[str, str], int]:
        """Dynamically calculates the vm type each region can use (both the source region and all destination regions)
        based on their quota limits and calculates the number of vms to launch in all regions by conservatively
        taking the minimum of all regions to stay consistent.

        :param src_region_tag: the source region tag (default: None)
        :type src_region_tag: Optional[str]
        :param dst_region_tags: a list of the destination region tags (defualt: None)
        :type dst_region_tags: Optional[List[str]]
        """

        # One of them has to provided
        # assert src_region_tag is not None or dst_region_tags is not None, "There needs to be at least one source or destination"
        src_tags = [src_region_tag] if src_region_tag is not None else []
        dst_tags = dst_region_tags if dst_region_tags is not None else []

        if src_region_tag:
            assert len(src_region_tag.split(":")) == 2, f"Source region tag {src_region_tag} must be in the form of `cloud_provider:region`"
        if dst_region_tags:
            assert (
                len(dst_region_tags[0].split(":")) == 2
            ), f"Destination region tag {dst_region_tags} must be in the form of `cloud_provider:region`"

        # do_parallel returns tuples of (region_tag, (vm_type, n_instances))
        vm_info = do_parallel(self._calculate_vm_types, src_tags + dst_tags)
        # Specifies the vm_type for each region
        vm_types = {v[0]: v[1][0] for v in vm_info}  # type: ignore
        # Taking the minimum so that we can use the same number of instances for both source and destination
        n_instances = min([self.n_instances] + [v[1][1] for v in vm_info])  # type: ignore
        return vm_types, n_instances


class UnicastDirectPlanner(Planner):
    # DO NOT USE THIS - broken for single-region transfers
    def __init__(self, n_instances: int, n_connections: int, transfer_config: TransferConfig, quota_limits_file: Optional[str] = None):
        super().__init__(transfer_config, quota_limits_file)
        self.n_instances = n_instances
        self.n_connections = n_connections

    def plan(self, jobs: List[TransferJob]) -> TopologyPlan:
        # make sure only single destination
        for job in jobs:
            assert len(job.dst_ifaces) == 1, f"DirectPlanner only support single destination jobs, got {len(job.dst_ifaces)}"

        src_region_tag = jobs[0].src_iface.region_tag()
        dst_region_tag = jobs[0].dst_ifaces[0].region_tag()

        assert len(src_region_tag.split(":")) == 2, f"Source region tag {src_region_tag} must be in the form of `cloud_provider:region`"
        assert (
            len(dst_region_tag.split(":")) == 2
        ), f"Destination region tag {dst_region_tag} must be in the form of `cloud_provider:region`"

        # jobs must have same sources and destinations
        for job in jobs[1:]:
            assert job.src_iface.region_tag() == src_region_tag, "All jobs must have same source region"
            assert job.dst_ifaces[0].region_tag() == dst_region_tag, "All jobs must have same destination region"

        plan = TopologyPlan(src_region_tag=src_region_tag, dest_region_tags=[dst_region_tag])

        # Dynammically calculate n_instances based on quota limits
        vm_types, n_instances = self._get_vm_type_and_instances(src_region_tag=src_region_tag, dst_region_tags=[dst_region_tag])

        # TODO: support on-sided transfers but not requiring VMs to be created in source/destination regions
        for i in range(n_instances):
            plan.add_gateway(src_region_tag, vm_types[src_region_tag])
            plan.add_gateway(dst_region_tag, vm_types[dst_region_tag])

        # ids of gateways in dst region
        dst_gateways = plan.get_region_gateways(dst_region_tag)

        src_program = GatewayProgram()
        dst_program = GatewayProgram()

        for job in jobs:
            src_bucket = job.src_iface.bucket()
            dst_bucket = job.dst_ifaces[0].bucket()

            # give each job a different partition id, so we can read/write to different buckets
            partition_id = job.uuid

            # source region gateway program
            obj_store_read = src_program.add_operator(
                GatewayReadObjectStore(src_bucket, src_region_tag, self.n_connections), partition_id=partition_id
            )
            mux_or = src_program.add_operator(GatewayMuxOr(), parent_handle=obj_store_read, partition_id=partition_id)
            for i in range(n_instances):
                src_program.add_operator(
                    GatewaySend(
                        target_gateway_id=dst_gateways[i].gateway_id,
                        region=src_region_tag,
                        num_connections=self.n_connections,
                        compress=True,
                        encrypt=True,
                    ),
                    parent_handle=mux_or,
                    partition_id=partition_id,
                )

            # dst region gateway program
            recv_op = dst_program.add_operator(GatewayReceive(decompress=True, decrypt=True), partition_id=partition_id)
            dst_program.add_operator(
                GatewayWriteObjectStore(dst_bucket, dst_region_tag, self.n_connections), parent_handle=recv_op, partition_id=partition_id
            )

            # update cost per GB
            plan.cost_per_gb += compute.CloudProvider.get_transfer_cost(src_region_tag, dst_region_tag)

        # set gateway programs
        plan.set_gateway_program(src_region_tag, src_program)
        plan.set_gateway_program(dst_region_tag, dst_program)

        return plan


class MulticastDirectPlanner(Planner):
    def __init__(self, n_instances: int, n_connections: int, transfer_config: TransferConfig, quota_limits_file: Optional[str] = None):
        super().__init__(transfer_config, quota_limits_file)
        self.n_instances = n_instances
        self.n_connections = n_connections

    def plan(self, jobs: List[TransferJob]) -> TopologyPlan:
        src_region_tag = jobs[0].src_iface.region_tag()
        dst_region_tags = [iface.region_tag() for iface in jobs[0].dst_ifaces]

        # jobs must have same sources and destinations
        for job in jobs[1:]:
            assert job.src_iface.region_tag() == src_region_tag, "All jobs must have same source region"
            assert [iface.region_tag() for iface in job.dst_ifaces] == dst_region_tags, "Add jobs must have same destination set"

        plan = TopologyPlan(src_region_tag=src_region_tag, dest_region_tags=dst_region_tags)

        # Dynammically calculate n_instances based on quota limits
        if src_region_tag.split(":")[0] == "test":
            vm_types = None
            n_instances = self.n_instances
        else:
            vm_types, n_instances = self._get_vm_type_and_instances(src_region_tag=src_region_tag, dst_region_tags=dst_region_tags)

        # TODO: support on-sided transfers but not requiring VMs to be created in source/destination regions
        for i in range(n_instances):
            plan.add_gateway(src_region_tag, vm_types[src_region_tag] if vm_types else None)
            for dst_region_tag in dst_region_tags:
                plan.add_gateway(dst_region_tag, vm_types[dst_region_tag] if vm_types else None)

        # initialize gateway programs per region
        dst_program = {dst_region: GatewayProgram() for dst_region in dst_region_tags}
        src_program = GatewayProgram()

        # iterate through all jobs
        for job in jobs:
            src_bucket = job.src_iface.bucket()
            src_region_tag = job.src_iface.region_tag()
            src_provider = src_region_tag.split(":")[0]

            # give each job a different partition id, so we can read/write to different buckets
            partition_id = job.uuid

            # source region gateway program
            obj_store_read = src_program.add_operator(
                GatewayReadObjectStore(src_bucket, src_region_tag, self.n_connections), partition_id=partition_id
            )

            # send to all destination
            mux_and = src_program.add_operator(GatewayMuxAnd(), parent_handle=obj_store_read, partition_id=partition_id)
            dst_prefixes = job.dst_prefixes
            for i in range(len(job.dst_ifaces)):
                dst_iface = job.dst_ifaces[i]
                dst_prefix = dst_prefixes[i]
                dst_region_tag = dst_iface.region_tag()
                dst_bucket = dst_iface.bucket()
                dst_gateways = plan.get_region_gateways(dst_region_tag)

                # special case where destination is same region as source
                if dst_region_tag == src_region_tag:
                    src_program.add_operator(
                        GatewayWriteObjectStore(dst_bucket, dst_region_tag, self.n_connections, key_prefix=dst_prefix),
                        parent_handle=mux_and,
                        partition_id=partition_id,
                    )
                    continue

                # can send to any gateway in region
                mux_or = src_program.add_operator(GatewayMuxOr(), parent_handle=mux_and, partition_id=partition_id)
                for i in range(n_instances):
                    private_ip = False
                    if dst_gateways[i].provider == "gcp" and src_provider == "gcp":
                        # print("Using private IP for GCP to GCP transfer", src_region_tag, dst_region_tag)
                        private_ip = True
                    src_program.add_operator(
                        GatewaySend(
                            target_gateway_id=dst_gateways[i].gateway_id,
                            region=dst_region_tag,
                            num_connections=int(self.n_connections / len(dst_gateways)),
                            private_ip=private_ip,
                            compress=self.transfer_config.use_compression,
                            encrypt=self.transfer_config.use_e2ee,
                        ),
                        parent_handle=mux_or,
                        partition_id=partition_id,
                    )

                # each gateway also recieves data from source
                recv_op = dst_program[dst_region_tag].add_operator(
                    GatewayReceive(decompress=self.transfer_config.use_compression, decrypt=self.transfer_config.use_e2ee),
                    partition_id=partition_id,
                )
                dst_program[dst_region_tag].add_operator(
                    GatewayWriteObjectStore(dst_bucket, dst_region_tag, self.n_connections, key_prefix=dst_prefix),
                    parent_handle=recv_op,
                    partition_id=partition_id,
                )

                # update cost per GB
                plan.cost_per_gb += compute.CloudProvider.get_transfer_cost(src_region_tag, dst_region_tag)

        # set gateway programs
        plan.set_gateway_program(src_region_tag, src_program)
        for dst_region_tag, program in dst_program.items():
            if dst_region_tag != src_region_tag:  # don't overwrite
                plan.set_gateway_program(dst_region_tag, program)
        return plan


class DirectPlannerSourceOneSided(MulticastDirectPlanner):
    """Planner that only creates VMs in the source region"""

    def plan(self, jobs: List[TransferJob]) -> TopologyPlan:
        src_region_tag = jobs[0].src_iface.region_tag()
        dst_region_tags = [iface.region_tag() for iface in jobs[0].dst_ifaces]
        # jobs must have same sources and destinations
        for job in jobs[1:]:
            assert job.src_iface.region_tag() == src_region_tag, "All jobs must have same source region"
            assert [iface.region_tag() for iface in job.dst_ifaces] == dst_region_tags, "Add jobs must have same destination set"

        plan = TopologyPlan(src_region_tag=src_region_tag, dest_region_tags=dst_region_tags)

        # Dynammically calculate n_instances based on quota limits
        vm_types, n_instances = self._get_vm_type_and_instances(src_region_tag=src_region_tag)

        # TODO: support on-sided transfers but not requiring VMs to be created in source/destination regions
        for i in range(n_instances):
            plan.add_gateway(src_region_tag, vm_types[src_region_tag])

        # initialize gateway programs per region
        src_program = GatewayProgram()

        # iterate through all jobs
        for job in jobs:
            src_bucket = job.src_iface.bucket()
            src_region_tag = job.src_iface.region_tag()
            src_provider = src_region_tag.split(":")[0]

            # give each job a different partition id, so we can read/write to different buckets
            partition_id = job.uuid

            # source region gateway program
            obj_store_read = src_program.add_operator(
                GatewayReadObjectStore(src_bucket, src_region_tag, self.n_connections), partition_id=partition_id
            )
            # send to all destination
            mux_and = src_program.add_operator(GatewayMuxAnd(), parent_handle=obj_store_read, partition_id=partition_id)
            dst_prefixes = job.dst_prefixes
            for i in range(len(job.dst_ifaces)):
                dst_iface = job.dst_ifaces[i]
                dst_prefix = dst_prefixes[i]
                dst_region_tag = dst_iface.region_tag()
                dst_bucket = dst_iface.bucket()
                dst_gateways = plan.get_region_gateways(dst_region_tag)

                # special case where destination is same region as source
                src_program.add_operator(
                    GatewayWriteObjectStore(dst_bucket, dst_region_tag, self.n_connections, key_prefix=dst_prefix),
                    parent_handle=mux_and,
                    partition_id=partition_id,
                )
                # update cost per GB
                plan.cost_per_gb += compute.CloudProvider.get_transfer_cost(src_region_tag, dst_region_tag)

        # set gateway programs
        plan.set_gateway_program(src_region_tag, src_program)
        return plan


class DirectPlannerDestOneSided(MulticastDirectPlanner):
    """Planner that only creates instances in the destination region"""

    def plan(self, jobs: List[TransferJob]) -> TopologyPlan:
        # only create in destination region
        src_region_tag = jobs[0].src_iface.region_tag()
        dst_region_tags = [iface.region_tag() for iface in jobs[0].dst_ifaces]
        # jobs must have same sources and destinations
        for job in jobs[1:]:
            assert job.src_iface.region_tag() == src_region_tag, "All jobs must have same source region"
            assert [iface.region_tag() for iface in job.dst_ifaces] == dst_region_tags, "Add jobs must have same destination set"

        plan = TopologyPlan(src_region_tag=src_region_tag, dest_region_tags=dst_region_tags)

        # Dynammically calculate n_instances based on quota limits
        vm_types, n_instances = self._get_vm_type_and_instances(dst_region_tags=dst_region_tags)

        # TODO: support on-sided transfers but not requiring VMs to be created in source/destination regions
        for i in range(n_instances):
            for dst_region_tag in dst_region_tags:
                plan.add_gateway(dst_region_tag, vm_types[dst_region_tag])

        # initialize gateway programs per region
        dst_program = {dst_region: GatewayProgram() for dst_region in dst_region_tags}

        # iterate through all jobs
        for job in jobs:
            src_bucket = job.src_iface.bucket()
            src_region_tag = job.src_iface.region_tag()
            src_provider = src_region_tag.split(":")[0]

            partition_id = job.uuid

            # send to all destination
            dst_prefixes = job.dst_prefixes
            for i in range(len(job.dst_ifaces)):
                dst_iface = job.dst_ifaces[i]
                dst_prefix = dst_prefixes[i]
                dst_region_tag = dst_iface.region_tag()
                dst_bucket = dst_iface.bucket()
                dst_gateways = plan.get_region_gateways(dst_region_tag)

                # source region gateway program
                obj_store_read = dst_program[dst_region_tag].add_operator(
                    GatewayReadObjectStore(src_bucket, src_region_tag, self.n_connections), partition_id=partition_id
                )

                dst_program[dst_region_tag].add_operator(
                    GatewayWriteObjectStore(dst_bucket, dst_region_tag, self.n_connections, key_prefix=dst_prefix),
                    parent_handle=obj_store_read,
                    partition_id=partition_id,
                )

                # update cost per GB
                plan.cost_per_gb += compute.CloudProvider.get_transfer_cost(src_region_tag, dst_region_tag)

        # set gateway programs
        for dst_region_tag, program in dst_program.items():
            plan.set_gateway_program(dst_region_tag, program)
        return plan



from skyplane.gateway.gateway_program import (
    GatewayProgram,
    GatewaySend,
    GatewayWriteLocal,
    GatewayWriteObjectStore,
    GatewayGenData,
    GatewayReadObjectStore,
)
from typing import List, Dict, Optional


class TopologyPlanGateway:

    """
    Represents a gateway in the topology plan.
    """

    def __init__(self, region_tag: str, gateway_id: str, gateway_vm: Optional[str]):
        self.region_tag = region_tag
        self.gateway_id = gateway_id
        self.gateway_vm = gateway_vm
        self.gateway_program = None

        # ip addresses
        self.private_ip_address = None
        self.public_ip_address = None

    @property
    def provider(self):
        """Get the provider of the gateway"""
        return self.region_tag.split(":")[0]

    @property
    def region(self):
        """Get the region of the gateway"""
        return self.region_tag.split(":")[1]

    @property
    def vm_type(self):
        """Get the vm type of the gateway"""
        return self.gateway_vm

    def set_private_ip_address(self, private_ip_address: str):
        """Set the IP address of the gateway (not determined until provisioning is complete)"""
        self.private_ip_address = private_ip_address

    def set_public_ip_address(self, public_ip_address: str):
        """Set the public IP address of the gateway (not determined until provisioning is complete)"""
        self.public_ip_address = public_ip_address

    def set_gateway_program(self, gateway_program: GatewayProgram):
        """Set the gateway program for the gateway"""
        self.gateway_program = gateway_program

    def write_operators(self) -> List[GatewayWriteObjectStore]:
        """Get all write operators in the gateway program"""
        # TODO: include other write operator types
        return [op for op in self.gateway_program.get_operators() if isinstance(op, GatewayWriteObjectStore)]


class TopologyPlan:
    """
    The TopologyPlan constains a list of gateway programs corresponding to each gateway in the dataplane.
    """

    def __init__(self, src_region_tag: str, dest_region_tags: List[str], cost_per_gb: float = 0.0):
        self.src_region_tag = src_region_tag
        self.dest_region_tags = dest_region_tags
        self.gateways = {}
        self.cost_per_gb = cost_per_gb

    @property
    def regions(self) -> List[str]:
        """Get all regions in the topology plan"""
        return list(set([gateway.region for gateway in self.gateways.values()]))

    @property
    def region_tags(self) -> List[str]:
        """Get all region tags in the topology plan"""
        return list(set([gateway.region_tag for gateway in self.gateways.values()]))

    def add_gateway(self, region_tag: str, vm_type: Optional[str] = None):
        """Create gateway in specified region"""
        gateway_id = region_tag + str(len([gateway for gateway in self.gateways.values() if gateway.region_tag == region_tag]))
        assert gateway_id not in self.gateways, f"Gateway id {gateway_id} in {self.gateways}"
        gateway = TopologyPlanGateway(region_tag, gateway_id, vm_type)
        self.gateways[gateway_id] = gateway
        return gateway

    def get_region_gateways(self, region_tag: str):
        """Get all gateways in a region"""
        return [gateway for gateway in self.gateways.values() if gateway.region_tag == region_tag]

    def get_gateways(self) -> List[TopologyPlanGateway]:
        """Get all gateways"""
        return list(self.gateways.values())

    def get_gateway(self, gateway_id: str) -> TopologyPlanGateway:
        return self.gateways[gateway_id]

    def set_gateway_program(self, region_tag: str, gateway_program: GatewayProgram):
        """Update all gateways in a region with specified gateway program"""
        for gateway in self.get_region_gateways(region_tag):
            gateway.set_gateway_program(gateway_program)

    def set_ip_addresses(self, gateway_id: str, private_ip_address: str, public_ip_address: str):
        """Set IP address of a gateway"""
        self.gateways[gateway_id].set_private_ip_address(private_ip_address)
        self.gateways[gateway_id].set_public_ip_address(public_ip_address)

    def generate_gateway_program(self, region_tag: str):
        """Generate gateway program for all gateways in a region"""
        # TODO: eventually let gateways in same region have different programs
        for gateway in self.get_region_gateways(region_tag):
            assert gateway.gateway_program is not None, f"Gateway program for {region_tag} has not been set"
            return gateway.gateway_program.to_json()

    def get_outgoing_paths(self, gateway_id: str):
        """Get all outgoing paths from a gateway"""
        outgoing_paths = {}
        for operator in self.gateways[gateway_id].gateway_program.get_operators():
            if isinstance(operator, GatewaySend):
                # get id of gateway that operator is sending to
                assert (
                    operator.target_gateway_id in self.gateways
                ), f"Gateway {operator.target_gateway_id} not found in gateway list {self.gateways}"
                outgoing_paths[operator.target_gateway_id] = operator.num_connections
        return outgoing_paths

    def get_gateway_program_json(self, gateway_id: str):
        """Get gateway program for a gateway"""
        return self.gateways[gateway_id].gateway_program.to_json()

    def get_gateway_info_json(self):
        """Return JSON mapping between gateway ids to public ip, public ip, provider, and region"""
        gateway_info = {}
        for gateway in self.gateways.values():
            gateway_info[gateway.gateway_id] = {
                "private_ip_address": gateway.private_ip_address,
                "public_ip_address": gateway.public_ip_address,
                "region": gateway.region,
                "provider": gateway.provider,
            }
        return gateway_info

    def sink_instances(self, region_tag: Optional[str] = None) -> Dict[str, List[TopologyPlanGateway]]:
        """Return list of gateways that have a sink operator (GatewayWriteObjectStore, GatewayWriteLocal)"""
        nodes = {}
        for gateway in self.gateways.values():
            for operator in gateway.gateway_program.get_operators():
                # dont include if wrong region
                if isinstance(operator, GatewayWriteObjectStore):
                    if region_tag is not None and operator.bucket_region != region_tag:
                        continue
                if isinstance(operator, GatewayWriteLocal):
                    if region_tag is not None and gateway.region_tag != region_tag:
                        continue

                if isinstance(operator, GatewayWriteObjectStore) or isinstance(operator, GatewayWriteLocal):
                    if gateway.region_tag not in nodes:
                        nodes[gateway.region_tag] = []
                    nodes[gateway.region_tag].append(gateway)
                    break
        return nodes

    def source_instances(self):
        """Return list of gateways that have a source operator (GatewayReadObjectStore, GatewayReadLocal, GatewayGenData)"""
        nodes = []
        for gateway in self.gateways.values():
            for operator in gateway.gateway_program.get_operators():
                if isinstance(operator, GatewayReadObjectStore) or isinstance(operator, GatewayGenData):
                    nodes.append(gateway)
                    break

        return nodes

    def per_region_count(self) -> Dict[str, int]:
        """Return number of gateways VMs per region"""
        counts = {}
        for node in self.get_gateways():
            counts[node.region_tag] = counts.get(node.region_tag, 0) + 1
        return counts

    def to_dict(self):
        return {gateway_id: gateway.gateway_program.to_dict() for gateway_id, gateway in self.gateways.items()}



from dataclasses import dataclass
from typing import Any, Iterator, List, Optional
from skyplane.obj_store.storage_interface import StorageInterface
import os


@dataclass
class LocalFile:
    """Defines file on local node."""

    path: str
    size: Optional[int] = None
    last_modified: Optional[str] = None
    file_format: Optional[str] = None

    @property
    def exists(self):
        raise NotImplementedError()

    def write_permissions(self):
        raise NotImplementedError()

    def real_path(self):
        raise NotImplementedError()


class FileSystemInterface(StorageInterface):
    def region_tag(self) -> str:
        return "local"

    def path(self) -> str:
        raise NotImplementedError()

    def list_files(self, prefix="") -> Iterator[LocalFile]:
        raise NotImplementedError()

    def get_file_size(self, file_name) -> int:
        raise NotImplementedError()

    def get_file_last_modified(self, file_name):
        raise NotImplementedError()

    def cache_file_locally(self, src_file_path, dst_file_path):
        # Incases where the data may be on a remote filesystem, we want to cache it locally
        raise NotImplementedError()

    def clear_cache(self):
        raise NotImplementedError()

    def delete_files(self, paths: List[str]):
        raise NotImplementedError()

    def initiate_multipart_upload(self, dst_object_name: str) -> str:
        raise ValueError("Multipart uploads not supported")

    def complete_multipart_upload(self, dst_object_name: str, upload_id: str, metadata: Optional[Any] = None) -> None:
        raise ValueError("Multipart uploads not supported")

    @staticmethod
    def create(fs: str, path: str, port: Optional[int] = None):
        if fs.startswith("hdfs"):
            from skyplane.obj_store.hdfs_interface import HDFSInterface

            return HDFSInterface(path, port)
        else:
            from skyplane.obj_store.posix_file_interface import POSIXInterface

            return POSIXInterface(path)



import json
import os
import threading
from collections import defaultdict, Counter
from datetime import datetime
from functools import partial
from datetime import datetime

import nacl.secret
import nacl.utils
import urllib3
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

from skyplane import compute
from skyplane.exceptions import GatewayContainerStartException
from skyplane.api.tracker import TransferProgressTracker, TransferHook
from skyplane.api.transfer_job import CopyJob, SyncJob, TransferJob
from skyplane.api.config import TransferConfig
from skyplane.planner.topology import TopologyPlan, TopologyPlanGateway
from skyplane.utils import logger
from skyplane.utils.definitions import gateway_docker_image, tmp_log_dir
from skyplane.utils.fn import PathLike, do_parallel

if TYPE_CHECKING:
    from skyplane.api.provisioner import Provisioner


class DataplaneAutoDeprovision:
    def __init__(self, dataplane: "Dataplane"):
        self.dataplane = dataplane

    def __enter__(self):
        return self.dataplane

    def __exit__(self, exc_type, exc_value, exc_tb):
        logger.fs.warning("Deprovisioning dataplane")
        self.dataplane.deprovision()


class Dataplane:
    """A Dataplane represents a concrete Skyplane network, including topology and VMs."""

    def __init__(
        self,
        clientid: str,
        topology: TopologyPlan,
        provisioner: "Provisioner",
        transfer_config: TransferConfig,
        log_dir: str,
        debug: bool = True,
    ):
        """
        :param clientid: the uuid of the local host to create the dataplane
        :type clientid: str
        :param topology: the calculated topology during the transfer
        :type topology: TopologyPlan
        :param provisioner: the provisioner to launch the VMs
        :type provisioner: Provisioner
        :param transfer_config: the configuration during the transfer
        :type transfer_config: TransferConfig
        """
        self.clientid = clientid
        self.topology = topology
        self.provisioner = provisioner
        self.transfer_config = transfer_config
        # disable for azure
        # TODO: remove this
        self.http_pool = urllib3.PoolManager(retries=urllib3.Retry(total=3))
        self.provisioning_lock = threading.Lock()
        self.provisioned = False
        self.log_dir = Path(log_dir)
        self.transfer_dir = tmp_log_dir / "transfer_logs" / datetime.now().strftime("%Y%m%d_%H%M%S")
        self.transfer_dir.mkdir(exist_ok=True, parents=True)

        # transfer logs
        self.transfer_dir = tmp_log_dir / "transfer_logs" / datetime.now().strftime("%Y%m%d_%H%M%S")
        self.transfer_dir.mkdir(exist_ok=True, parents=True)
        self.debug = debug

        # pending tracker tasks
        self.pending_transfers: List[TransferProgressTracker] = []
        self.bound_nodes: Dict[TopologyPlanGateway, compute.Server] = {}

    def _start_gateway(
        self,
        gateway_docker_image: str,
        gateway_node: TopologyPlanGateway,
        gateway_server: compute.Server,
        gateway_log_dir: Optional[PathLike],
        authorize_ssh_pub_key: Optional[str] = None,
        e2ee_key_bytes: Optional[str] = None,
    ):
        # map outgoing ports
        setup_args = {}
        for gateway_id, n_conn in self.topology.get_outgoing_paths(gateway_node.gateway_id).items():
            node = self.topology.get_gateway(gateway_id)
            # use private ips for gcp to gcp connection
            src_provider, dst_provider = gateway_node.region.split(":")[0], node.region.split(":")[0]
            if src_provider == dst_provider and src_provider == "gcp":
                setup_args[self.bound_nodes[node].private_ip()] = n_conn
            else:
                setup_args[self.bound_nodes[node].public_ip()] = n_conn
        logger.fs.debug(f"[Dataplane._start_gateway] Setup args for {gateway_node}: {setup_args}")

        if gateway_log_dir:
            gateway_server.init_log_files(gateway_log_dir)
        if authorize_ssh_pub_key:
            gateway_server.copy_public_key(authorize_ssh_pub_key)

        # write gateway programs
        gateway_program_filename = Path(f"{gateway_log_dir}/gateway_program_{gateway_node.gateway_id}.json")
        with open(gateway_program_filename, "w") as f:
            f.write(gateway_node.gateway_program.to_json())

        # start gateway
        gateway_server.start_gateway(
            # setup_args,
            gateway_docker_image=gateway_docker_image,
            gateway_program_path=str(gateway_program_filename),
            gateway_info_path=f"{gateway_log_dir}/gateway_info.json",
            e2ee_key_bytes=e2ee_key_bytes,  # TODO: remove
            use_bbr=self.transfer_config.use_bbr,  # TODO: remove
            use_compression=self.transfer_config.use_compression,
            use_socket_tls=self.transfer_config.use_socket_tls,
        )

    def provision(
        self,
        allow_firewall: bool = True,
        gateway_docker_image: str = os.environ.get("SKYPLANE_DOCKER_IMAGE", gateway_docker_image()),
        authorize_ssh_pub_key: Optional[str] = None,
        max_jobs: int = 16,
        spinner: bool = False,
    ):
        """
        Provision the transfer gateways.

        :param allow_firewall: whether to apply firewall rules in the gatweway network (default: True)
        :type allow_firewall: bool
        :param gateway_docker_image: Docker image token in github
        :type gateway_docker_image: str
        :param authorize_ssh_pub_key: authorization ssh key to the remote gateways
        :type authorize_ssh_pub_key: str
        :param max_jobs: maximum number of provision jobs to launch concurrently (default: 16)
        :type max_jobs: int
        :param spinner: whether to show the spinner during the job (default: False)its to determine how many instances to create in each region
        :type spinner: bool
        """
        # TODO: support on-sided transfers but not requiring VMs to be created in source/destination regions
        with self.provisioning_lock:
            if self.provisioned:
                logger.error("Cannot provision dataplane, already provisioned!")
                return
            is_aws_used = any(n.region_tag.startswith("aws:") for n in self.topology.get_gateways())
            is_azure_used = any(n.region_tag.startswith("azure:") for n in self.topology.get_gateways())
            is_gcp_used = any(n.region_tag.startswith("gcp:") for n in self.topology.get_gateways())
            is_ibmcloud_used = any(n.region_tag.startswith("ibmcloud:") for n in self.topology.get_gateways())

            # create VMs from the topology
            for node in self.topology.get_gateways():
                cloud_provider, region = node.region_tag.split(":")
                assert (
                    cloud_provider != "cloudflare"
                ), f"Cannot create VMs in certain cloud providers: check planner output {self.topology.to_dict()}"
                self.provisioner.add_task(
                    cloud_provider=cloud_provider,
                    region=region,
                    vm_type=node.vm_type or getattr(self.transfer_config, f"{cloud_provider}_instance_class"),
                    spot=getattr(self.transfer_config, f"{cloud_provider}_use_spot_instances"),
                    autoterminate_minutes=self.transfer_config.autoterminate_minutes,
                )

            # initialize clouds
            self.provisioner.init_global(aws=is_aws_used, azure=is_azure_used, gcp=is_gcp_used, ibmcloud=is_ibmcloud_used)

            # provision VMs
            uuids = self.provisioner.provision(
                authorize_firewall=allow_firewall,
                max_jobs=max_jobs,
                spinner=spinner,
            )

            # bind VMs to nodes
            servers = [self.provisioner.get_node(u) for u in uuids]
            servers_by_region = defaultdict(list)
            for s in servers:
                servers_by_region[s.region_tag].append(s)
            for node in self.topology.get_gateways():
                instance = servers_by_region[node.region_tag].pop()
                self.bound_nodes[node] = instance

                # set ip addresses (for gateway program generation)
                self.topology.set_ip_addresses(node.gateway_id, self.bound_nodes[node].private_ip(), self.bound_nodes[node].public_ip())

            logger.fs.debug(f"[Dataplane.provision] bound_nodes = {self.bound_nodes}")
            gateway_bound_nodes = self.bound_nodes.copy()

            # start gateways
            self.provisioned = True

        # todo: move server.py:start_gateway here
        logger.fs.info(f"Using docker image {gateway_docker_image}")
        e2ee_key_bytes = nacl.utils.random(nacl.secret.SecretBox.KEY_SIZE)

        # create gateway logging dir
        gateway_program_dir = f"{self.log_dir}/programs"
        Path(gateway_program_dir).mkdir(exist_ok=True, parents=True)
        logger.fs.info(f"Writing gateway programs to {gateway_program_dir}")

        # write gateway info file
        gateway_info_path = f"{gateway_program_dir}/gateway_info.json"
        with open(gateway_info_path, "w") as f:
            json.dump(self.topology.get_gateway_info_json(), f, indent=4)
        logger.fs.info(f"Writing gateway info to {gateway_info_path}")

        # start gateways in parallel
        jobs = []
        for node, server in gateway_bound_nodes.items():
            jobs.append(
                partial(self._start_gateway, gateway_docker_image, node, server, gateway_program_dir, authorize_ssh_pub_key, e2ee_key_bytes)
            )
        logger.fs.debug(f"[Dataplane.provision] Starting gateways on {len(jobs)} servers")
        try:
            do_parallel(lambda fn: fn(), jobs, n=-1, spinner=spinner, spinner_persist=spinner, desc="Starting gateway container on VMs")
        except Exception as e:
            self.copy_gateway_logs()
            raise GatewayContainerStartException(f"Error starting gateways. Please check gateway logs {self.transfer_dir}")

    def copy_gateway_logs(self):
        # copy logs from all gateways in parallel
        def copy_log(instance):
            out_file = self.transfer_dir / f"gateway_{instance.uuid()}.stdout"
            err_file = self.transfer_dir / f"gateway_{instance.uuid()}.stderr"
            logger.fs.info(f"[Dataplane.copy_gateway_logs] Copying logs from {instance.uuid()}: {out_file}")
            instance.run_command("sudo docker logs -t skyplane_gateway 2> /tmp/gateway.stderr > /tmp/gateway.stdout")
            instance.download_file("/tmp/gateway.stdout", out_file)
            instance.download_file("/tmp/gateway.stderr", err_file)

        do_parallel(copy_log, self.bound_nodes.values(), n=-1)

    def deprovision(self, max_jobs: int = 64, spinner: bool = False):
        """
        Deprovision the remote gateways

        :param max_jobs: maximum number of jobs to deprovision the remote gateways (default: 64)
        :type max_jobs: int
        :param spinner: Whether to show the spinner during the job (default: False)
        :type spinner: bool
        """
        with self.provisioning_lock:
            if self.debug and self.provisioned:
                logger.fs.info(f"Copying gateway logs to {self.transfer_dir}")
                self.copy_gateway_logs()

            if not self.provisioned:
                logger.fs.warning("Attempting to deprovision dataplane that is not provisioned, this may be from auto_deprovision.")
            # wait for tracker tasks
            try:
                for task in self.pending_transfers:
                    logger.fs.warning(f"Before deprovisioning, waiting for jobs to finish: {list(task.jobs.keys())}")
                    task.join()
            except KeyboardInterrupt:
                logger.warning("Interrupted while waiting for transfers to finish, deprovisioning anyway.")
                raise
            finally:
                self.provisioner.deprovision(
                    max_jobs=max_jobs,
                    spinner=spinner,
                )
                self.provisioned = False

    def check_error_logs(self) -> Dict[str, List[str]]:
        """Get the error log from remote gateways if there is any error."""

        def get_error_logs(args):
            _, instance = args
            reply = self.http_pool.request("GET", f"{instance.gateway_api_url}/api/v1/errors")
            if reply.status != 200:
                raise Exception(f"Failed to get error logs from gateway instance {instance.instance_name()}: {reply.data.decode('utf-8')}")
            return json.loads(reply.data.decode("utf-8"))["errors"]

        errors: Dict[str, List[str]] = {}
        for (_, instance), result in do_parallel(get_error_logs, self.bound_nodes.items(), n=8):
            errors[instance] = result
        return errors

    def auto_deprovision(self) -> DataplaneAutoDeprovision:
        """Returns a context manager that will automatically call deprovision upon exit."""
        return DataplaneAutoDeprovision(self)

    def source_gateways(self) -> List[compute.Server]:
        """Returns a list of source gateway nodes"""
        return [self.bound_nodes[n] for n in self.topology.source_instances()] if self.provisioned else []

    def sink_gateways(self, region_tag: Optional[str] = None) -> Dict[str, List[compute.Server]]:
        """Returns a list of sink gateway nodes"""
        return (
            {region: [self.bound_nodes[n] for n in nodes] for region, nodes in self.topology.sink_instances(region_tag).items()}
            if self.provisioned
            else {}
        )

    def run_async(self, jobs: List[TransferJob], hooks: Optional[TransferHook] = None) -> TransferProgressTracker:
        """Start the transfer asynchronously. The main thread will not be blocked.

        :param hooks: Tracks the status of the transfer
        :type hooks: TransferHook
        """
        if not self.provisioned:
            logger.error("Dataplane must be pre-provisioned. Call dataplane.provision() before starting a transfer")
        tracker = TransferProgressTracker(self, jobs, self.transfer_config, hooks)
        self.pending_transfers.append(tracker)
        tracker.start()
        logger.fs.info(f"[SkyplaneClient] Started async transfer with {len(jobs)} jobs")
        return tracker

    def run(self, jobs: List[TransferJob], hooks: Optional[TransferHook] = None):
        """Start the transfer in the main thread. Wait until the transfer is complete.

        :param hooks: Tracks the status of the transfer
        :type hooks: TransferHook
        """
        tracker = self.run_async(jobs, hooks)
        logger.fs.debug(f"[SkyplaneClient] Waiting for transfer to complete")
        tracker.join()



import json
import time
import time
import typer
import math
import queue
import sys
import threading
import time
import uuid
from abc import ABC
from collections import defaultdict
from dataclasses import dataclass, field
from queue import Queue

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Generator, List, Optional, Tuple, TypeVar, Dict

from abc import ABC, abstractmethod

import urllib3
from rich import print as rprint
from functools import partial

from skyplane import exceptions
from skyplane.api.config import TransferConfig
from skyplane.chunk import Chunk, ChunkRequest
from skyplane.obj_store.storage_interface import StorageInterface
from skyplane.obj_store.object_store_interface import ObjectStoreObject, ObjectStoreInterface
from skyplane.utils import logger
from skyplane.utils.definitions import MB
from skyplane.utils.fn import do_parallel
from skyplane.utils.path import parse_path
from skyplane.utils.retry import retry_backoff

if TYPE_CHECKING:
    from skyplane.api.dataplane import Dataplane

T = TypeVar("T")


@dataclass
class TransferPair:
    "Represents transfer pair between source and destination"

    def __init__(self, src_obj: ObjectStoreObject, dst_objs: Dict[str, ObjectStoreObject], dst_key: str):
        self.src_obj = src_obj
        self.dst_objs = dst_objs
        self.dst_key = dst_key  # shared destination key across all chunks (differnt prefixes)


@dataclass
class GatewayMessage:
    def __init__(self, chunk: Optional[Chunk] = None, upload_id_mapping: Optional[Dict[str, Tuple[str, str]]] = None):
        self.chunk = chunk
        # TODO: currently, the mapping ID is per-region, however this should be per-bucket, as there may be multiple
        # target buckets in the same region
        self.upload_id_mapping = upload_id_mapping  # region_tag: (upload_id, dst_key)


class Chunker:
    """class that chunks the original files and makes the chunk requests"""

    def __init__(
        self,
        src_iface: StorageInterface,
        dst_ifaces: List[StorageInterface],
        transfer_config: TransferConfig,
        concurrent_multipart_chunk_threads: Optional[int] = 64,
        num_partitions: Optional[int] = 1,
    ):
        """
        :param src_iface: source object store interface
        :type src_iface: StorageInterface
        :param dst_iface: destination object store interface
        :type dst_ifaces: List[StorageInterface]
        :param transfer_config: the configuration during the transfer
        :type transfer_config: TransferConfig
        :param concurrent_multipart_chunk_threads: the maximum number of concurrent threads that dispatch multipart chunk requests (default: 64)
        :type concurrent_multipart_chunk_threads: int
        """
        self.src_iface = src_iface
        self.dst_ifaces = dst_ifaces
        self.transfer_config = transfer_config
        self.multipart_upload_requests = []
        self.concurrent_multipart_chunk_threads = concurrent_multipart_chunk_threads
        self.num_partitions = num_partitions

    def _run_multipart_chunk_thread(
        self,
        exit_event: threading.Event,
        in_queue: "Queue[TransferPair]",
        out_queue_chunks: "Queue[GatewayMessage]",
    ):
        """Chunks large files into many small chunks."""
        while not exit_event.is_set():
            try:
                transfer_pair = in_queue.get(block=False, timeout=0.1)
            except queue.Empty:
                continue

            src_object = transfer_pair.src_obj
            dest_objects = transfer_pair.dst_objs
            dest_key = transfer_pair.dst_key
            if isinstance(self.src_iface, ObjectStoreInterface):
                mime_type = self.src_iface.get_obj_mime_type(src_object.key)
                # create multipart upload request per destination
                upload_id_mapping = {}
                for dest_iface in self.dst_ifaces:
                    dest_object = dest_objects[dest_iface.region_tag()]
                    upload_id = dest_iface.initiate_multipart_upload(dest_object.key, mime_type=mime_type)
                    # print(f"Created upload id for key {dest_object.key} with upload id {upload_id} for bucket {dest_iface.bucket_name}")
                    # store mapping between key and upload id for each region
                    upload_id_mapping[dest_iface.region_tag()] = (src_object.key, upload_id)
                out_queue_chunks.put(GatewayMessage(upload_id_mapping=upload_id_mapping))  # send to output queue

                # get source and destination object and then compute number of chunks
                chunk_size_bytes = int(self.transfer_config.multipart_chunk_size_mb * MB)
                num_chunks = math.ceil(src_object.size / chunk_size_bytes)
                if num_chunks > self.transfer_config.multipart_max_chunks:
                    chunk_size_bytes = int(src_object.size / self.transfer_config.multipart_max_chunks)
                    chunk_size_bytes = math.ceil(chunk_size_bytes / MB) * MB  # round to next largest mb
                    num_chunks = math.ceil(src_object.size / chunk_size_bytes)

                assert num_chunks * chunk_size_bytes >= src_object.size
                # create chunks
                offset = 0
                part_num = 1
                parts = []
                for _ in range(num_chunks):
                    file_size_bytes = min(chunk_size_bytes, src_object.size - offset)
                    assert file_size_bytes > 0, f"file size <= 0 {file_size_bytes}"
                    chunk = Chunk(
                        src_key=src_object.key,
                        dest_key=dest_key,  # dest_object.key, # TODO: upload basename (no prefix)
                        chunk_id=uuid.uuid4().hex,
                        file_offset_bytes=offset,
                        partition_id=str(part_num % self.num_partitions),
                        chunk_length_bytes=file_size_bytes,
                        part_number=part_num,
                        # upload_id=upload_id,
                        multi_part=True,
                    )
                    assert upload_id is not None, f"Upload id cannot be None for multipart upload for {src_object.key}"
                    assert part_num is not None, f"Partition cannot be none {part_num}"
                    offset += file_size_bytes
                    parts.append(part_num)
                    part_num += 1
                    out_queue_chunks.put(GatewayMessage(chunk=chunk))

                # store multipart ids
                for dest_iface in self.dst_ifaces:
                    bucket = dest_iface.bucket()
                    region = dest_iface.region_tag()
                    dest_object = dest_objects[region]
                    _, upload_id = upload_id_mapping[region]

                    metadata = None
                    # Convert parts to base64 and store mime_type if destination interface is AzureBlobInterface
                    if dest_iface.provider == "azure":
                        from skyplane.obj_store.azure_blob_interface import AzureBlobInterface

                        block_ids = list(map(lambda part_num: AzureBlobInterface.id_to_base64_encoding(part_num, dest_object.key), parts))
                        metadata = (block_ids, mime_type)

                    self.multipart_upload_requests.append(
                        dict(upload_id=upload_id, key=dest_object.key, parts=parts, region=region, bucket=bucket, metadata=metadata)
                    )
            else:
                mime_type = None
                raise NotImplementedError("Multipart not implement for non-object store interfaces")

    # def to_chunk_requests(self, gen_in: Generator[Chunk, None, None]) -> Generator[ChunkRequest, None, None]:
    #    """Converts a generator of chunks to a generator of chunk requests.

    #    :param gen_in: generator that generates chunk requests
    #    :type gen_in: Generator
    #    """
    #    src_region = self.src_iface.region_tag()
    #    src_bucket = self.src_iface.bucket()
    #    for chunk in gen_in:
    #        yield ChunkRequest(
    #            chunk=chunk,
    #            src_region=src_region,
    #            #dst_region=dest_region,
    #            src_object_store_bucket=src_bucket,
    #            #dst_object_store_bucket=dest_bucket,
    #            src_type="object_store",
    #            #dst_type="object_store",
    #        )

    @staticmethod
    def map_object_key_prefix(source_prefix: str, source_key: str, dest_prefix: str, recursive: bool = False):
        """
        map_object_key_prefix computes the mapping of a source key in a bucket prefix to the destination.
        Users invoke a transfer via the CLI; aws s3 cp s3://bucket/source_prefix s3://bucket/dest_prefix.
        The CLI will query the object store for all objects in the source prefix and map them to the
        destination prefix using this function.

        :param source_prefix: source bucket folder prefix
        :type source_prefix: string
        :param source_key: source file key to map in the folder prefix
        :type source_key: string
        :param destination_prefix: destination bucket folder prefix
        :type destination_prefix: string
        :param recursive: whether to copy all the objects matching the pattern (default: False)
        :type recursive: bool
        """
        join = lambda prefix, fname: prefix + fname if prefix.endswith("/") else prefix + "/" + fname
        src_fname = source_key.split("/")[-1] if "/" in source_key and not source_key.endswith("/") else source_key
        if not recursive:
            if source_key == source_prefix:
                if dest_prefix == "" or dest_prefix == "/":
                    return src_fname
                elif dest_prefix[-1] == "/":
                    return dest_prefix + src_fname
                else:
                    return dest_prefix
            else:
                # todo: don't print output here
                rprint(f"\n:x: [bold red]In order to transfer objects using a prefix, you must use the --recursive or -r flag.[/bold red]")
                rprint(f"[yellow]If you meant to transfer a single object, pass the full source object key.[/yellow]")
                rprint(f"[bright_black]Try running: [bold]skyplane {' '.join(sys.argv[1:])} --recursive[/bold][/bright_black]")
                raise exceptions.MissingObjectException("Encountered a recursive transfer without the --recursive flag.") from None
        else:
            if source_prefix == "" or source_prefix == "/":
                if dest_prefix == "" or dest_prefix == "/":
                    return source_key
                else:
                    return join(dest_prefix, source_key)
            else:
                # catch special case: map_object_key_prefix("foo", "foobar/baz.txt", "", recursive=True)
                if not source_key.startswith(source_prefix + "/" if not source_prefix.endswith("/") else source_prefix):
                    rprint(f"\n:x: [bold red]The source key {source_key} does not start with the source prefix {source_prefix}[/bold red]")
                    raise exceptions.MissingObjectException(f"Source key {source_key} does not start with source prefix {source_prefix}")
                if dest_prefix == "" or dest_prefix == "/":
                    return source_key[len(source_prefix) :]
                else:
                    src_path_after_prefix = source_key[len(source_prefix) :]
                    src_path_after_prefix = src_path_after_prefix[1:] if src_path_after_prefix.startswith("/") else src_path_after_prefix
                    return join(dest_prefix, src_path_after_prefix)

    def transfer_pair_generator(
        self,
        src_prefix: str,
        dst_prefixes: List[str],
        recursive: bool,
        prefilter_fn: Optional[Callable[[ObjectStoreObject], bool]] = None,  # TODO: change to StorageObject
    ) -> Generator[TransferPair, None, None]:
        """Query source region and return list of objects to transfer.

        :param src_prefix: source bucket folder prefix
        :type src_prefix: string
        :param dst_prefix: destination bucket folder prefix
        :type dst_prefix: string
        :param recursive: if true, will copy objects at folder prefix recursively
        :type recursive: bool
        :param prefilter_fn: filters out objects whose prefixes do not match the filter function (default: None)
        :type prefilter_fn: Callable[[ObjectStoreObject], bool]
        """
        if not isinstance(self.src_iface, ObjectStoreInterface) or not all(
            [isinstance(dst_iface, ObjectStoreInterface) for dst_iface in self.dst_ifaces]
        ):
            raise NotImplementedError("TransferPair only supports object store interfaces")

        if not self.src_iface.bucket_exists():
            raise exceptions.MissingBucketException(f"Source bucket {self.src_iface.path()} does not exist or is not readable.")
        for dst_iface in self.dst_ifaces:
            if not dst_iface.bucket_exists():
                raise exceptions.MissingBucketException(f"Destination bucket {dst_iface.path()} does not exist or is not readable.")

        # query all source region objects
        logger.fs.debug(f"Querying objects in {self.src_iface.path()}")
        n_objs = 0
        for obj in self.src_iface.list_objects(src_prefix):
            if prefilter_fn is None or prefilter_fn(obj):
                # collect list of destination objects
                dest_objs = {}
                dest_keys = []
                for i in range(len(self.dst_ifaces)):
                    dst_iface = self.dst_ifaces[i]
                    dst_prefix = dst_prefixes[i]
                    dest_provider, dest_region = dst_iface.region_tag().split(":")
                    try:
                        dest_key = self.map_object_key_prefix(src_prefix, obj.key, dst_prefix, recursive=recursive)
                        assert (
                            dest_key[: len(dst_prefix)] == dst_prefix
                        ), f"Destination key {dest_key} does not start with destination prefix {dst_prefix}"
                        dest_keys.append(dest_key[len(dst_prefix) :])
                    except exceptions.MissingObjectException as e:
                        logger.fs.exception(e)
                        raise e from None

                    if dest_provider == "aws":
                        from skyplane.obj_store.s3_interface import S3Object

                        dest_obj = S3Object(provider=dest_provider, bucket=dst_iface.bucket(), key=dest_key)
                    elif dest_provider == "azure":
                        from skyplane.obj_store.azure_blob_interface import AzureBlobObject

                        dest_obj = AzureBlobObject(provider=dest_provider, bucket=dst_iface.bucket(), key=dest_key)
                    elif dest_provider == "gcp":
                        from skyplane.obj_store.gcs_interface import GCSObject

                        dest_obj = GCSObject(provider=dest_provider, bucket=dst_iface.bucket(), key=dest_key)
                    elif dest_provider == "cloudflare":
                        from skyplane.obj_store.r2_interface import R2Object

                        dest_obj = R2Object(provider=dest_provider, bucket=dst_iface.bucket(), key=dest_key)
                    else:
                        raise ValueError(f"Invalid dest_region {dest_region}, unknown provider")
                    dest_objs[dst_iface.region_tag()] = dest_obj

                # assert that all destinations share the same post-fix key
                assert len(list(set(dest_keys))) == 1, f"Destination keys {dest_keys} do not match"
                n_objs += 1
                yield TransferPair(src_obj=obj, dst_objs=dest_objs, dst_key=dest_keys[0])

        if n_objs == 0:
            logger.error("Specified object does not exist.\n")
            raise exceptions.MissingObjectException(f"No objects were found in the specified prefix")

    def chunk(self, transfer_pair_generator: Generator[TransferPair, None, None]) -> Generator[GatewayMessage, None, None]:
        """Break transfer list into chunks.

        :param transfer_pair_generator: generator of pairs of objects to transfer
        :type transfer_pair_generator: Generator
        """
        multipart_send_queue: Queue[TransferPair] = Queue()
        multipart_chunk_queue: Queue[GatewayMessage] = Queue()
        multipart_exit_event = threading.Event()
        multipart_chunk_threads = []

        # start chunking threads
        if self.transfer_config.multipart_enabled:
            for _ in range(self.concurrent_multipart_chunk_threads):
                t = threading.Thread(
                    target=self._run_multipart_chunk_thread,
                    args=(multipart_exit_event, multipart_send_queue, multipart_chunk_queue),
                    daemon=False,
                )
                t.start()
                multipart_chunk_threads.append(t)

        # begin chunking loop
        for transfer_pair in transfer_pair_generator:
            # print("transfer_pair", transfer_pair.src_obj.key, transfer_pair.dst_objs)
            src_obj = transfer_pair.src_obj
            if self.transfer_config.multipart_enabled and src_obj.size > self.transfer_config.multipart_threshold_mb * MB:
                multipart_send_queue.put(transfer_pair)
            else:
                if transfer_pair.src_obj.size == 0:
                    logger.fs.debug(f"Skipping empty object {src_obj.key}")
                    continue
                yield GatewayMessage(
                    chunk=Chunk(
                        src_key=src_obj.key,
                        dest_key=transfer_pair.dst_key,  # TODO: get rid of dest_key, and have write object have info on prefix  (or have a map here)
                        chunk_id=uuid.uuid4().hex,
                        chunk_length_bytes=transfer_pair.src_obj.size,
                        partition_id=str(0),  # TODO: fix this to distribute across multiple partitions
                    )
                )

            if self.transfer_config.multipart_enabled:
                # drain multipart chunk queue and yield with updated chunk IDs
                while not multipart_chunk_queue.empty():
                    yield multipart_chunk_queue.get()

        if self.transfer_config.multipart_enabled:
            # wait for processing multipart requests to finish
            logger.fs.debug("Waiting for multipart threads to finish")
            # while not multipart_send_queue.empty():
            # TODO: may be an issue waiting for this in case of force-quit
            while not multipart_send_queue.empty():
                logger.fs.debug(f"Remaining in multipart queue: sent {multipart_send_queue.qsize()}")
                time.sleep(0.1)
            # send sentinel to all threads
            multipart_exit_event.set()
            for thread in multipart_chunk_threads:
                thread.join()

            # drain multipart chunk queue and yield with updated chunk IDs
            while not multipart_chunk_queue.empty():
                yield multipart_chunk_queue.get()

    @staticmethod
    def batch_generator(gen_in: Generator[T, None, None], batch_size: int) -> Generator[List[T], None, None]:
        """Batches generator, while handling StopIteration

        :param gen_in: generator that generates chunk requests
        :type gen_in: Generator
        """
        batch = []
        for item in gen_in:
            batch.append(item)
            if len(batch) == batch_size:
                yield batch
                batch = []
        if len(batch) > 0:
            yield batch

    @staticmethod
    def prefetch_generator(gen_in: Generator[T, None, None], buffer_size: int) -> Generator[T, None, None]:
        """
        Prefetches from generator while handing StopIteration to ensure items yield immediately.
        Start a thread to prefetch items from the generator and put them in a queue. Upon StopIteration,
        the thread will add a sentinel value to the queue.

        :param gen_in: generator that generates chunk requests
        :type gen_in: Generator
        :param buffer_size: maximum size of the buffer to temporarily store the generators
        :type buffer_size: int
        """
        sentinel = object()
        queue = Queue(maxsize=buffer_size)

        def prefetch():
            for item in gen_in:
                queue.put(item)
            queue.put(sentinel)

        thread = threading.Thread(target=prefetch, daemon=True)
        thread.start()

        while True:
            item = queue.get()
            if item is sentinel:
                break
            yield item

    @staticmethod
    def tail_generator(gen_in: Generator[T, None, None], out_list: List[T]) -> Generator[T, None, None]:
        """Tails generator while handling StopIteration

        :param gen_in: generator that generates chunk requests
        :type gen_in: Generator
        :param out_list: list of tail generators
        :type out_list: List
        """
        for item in gen_in:
            out_list.append(item)
            yield item


@dataclass
class TransferJob(ABC):
    """
    transfer job with transfer configurations

    :param src_path: source full path
    :type src_path: str
    :param dst_paths: destination full path
    :type dst_paths: str
    :param recursive: if true, will transfer objects at folder prefix recursively (default: False)
    :type recursive: bool
    :param requester_pays: if set, will support requester pays buckets. (default: False)
    :type requester_pays: bool
    :param uuid: the uuid of one single transfer job
    :type uuid: str
    """

    # @abstractmethod
    def __init__(
        self,
        src_path: str,
        dst_paths: List[str] or str,
        recursive: bool = False,
        requester_pays: bool = False,
        job_id: Optional[str] = None,
    ):
        self.src_path = src_path
        self.dst_paths = dst_paths
        self.recursive = recursive
        self.requester_pays = requester_pays
        if job_id is None:
            self.uuid = str(uuid.uuid4())
        else:
            self.uuid = job_id

    @property
    def transfer_type(self) -> str:
        if isinstance(self.dst_paths, str):
            return "unicast"
        else:
            return "multicast"

    @property
    def src_prefix(self) -> Optional[str]:
        """Return the source prefix"""
        if not hasattr(self, "_src_prefix"):
            self._src_prefix = parse_path(self.src_path)[2]
        return self._src_prefix

    @property
    def src_iface(self) -> StorageInterface:
        """Return the source object store interface"""
        if not hasattr(self, "_src_iface"):
            provider_src, bucket_src, _ = parse_path(self.src_path)
            self._src_iface = ObjectStoreInterface.create(f"{provider_src}:infer", bucket_src)
            if self.requester_pays:
                self._src_iface.set_requester_bool(True)
        return self._src_iface

    @property
    def dst_prefixes(self) -> List[str]:
        """Return the destination prefix"""
        if not hasattr(self, "_dst_prefix"):
            if self.transfer_type == "unicast":
                self._dst_prefix = [str(parse_path(self.dst_paths[0])[2])]
            else:
                self._dst_prefix = [str(parse_path(path)[2]) for path in self.dst_paths]
        return self._dst_prefix

    @property
    def dst_ifaces(self) -> List[StorageInterface]:
        """Return the destination object store interface"""
        if not hasattr(self, "_dst_iface"):
            if self.transfer_type == "unicast":
                provider_dst, bucket_dst, _ = parse_path(self.dst_paths[0])
                self._dst_ifaces = [StorageInterface.create(f"{provider_dst}:infer", bucket_dst)]
            else:
                self._dst_ifaces = []
                for path in self.dst_paths:
                    provider_dst, bucket_dst, _ = parse_path(path)
                    self._dst_ifaces.append(StorageInterface.create(f"{provider_dst}:infer", bucket_dst))
        return self._dst_ifaces

    def dispatch(self, dataplane: "Dataplane", **kwargs) -> Generator[Chunk, None, None]:
        """Dispatch transfer job to specified gateways."""
        raise NotImplementedError("Dispatch not implemented")

    def finalize(self):
        """Complete the multipart upload requests"""
        raise NotImplementedError("Finalize not implemented")

    def verify(self):
        """Verifies the transfer completed, otherwise raises TransferFailedException."""
        raise NotImplementedError("Verify not implemented")

    def size_gb(self):
        """Return the size of the transfer in GB"""
        raise NotImplementedError("Size not implemented")

    @classmethod
    def _pre_filter_fn(cls, obj: ObjectStoreObject) -> bool:
        """Optionally filter source objects before they are transferred.

        :meta private:
        :param obj: source object to be transferred
        :type obj: ObjectStoreObject
        """
        return True


@dataclass
class CopyJob(TransferJob):
    """copy job that copies the source objects to the destination

    :param transfer_list: transfer list for later verification
    :type transfer_list: list
    :param multipart_transfer_list: multipart transfer list for later verification
    :type multipart_transfer_list: list
    """

    def __init__(
        self,
        src_path: str,
        dst_paths: List[str] or str,
        recursive: bool = False,
        requester_pays: bool = False,
        job_id: Optional[str] = None,
    ):
        super().__init__(src_path, dst_paths, recursive, requester_pays, job_id)
        self.transfer_list = []
        self.multipart_transfer_list = []

    @property
    def http_pool(self):
        """http connection pool"""
        if not hasattr(self, "_http_pool"):
            timeout = urllib3.util.Timeout(connect=10.0, read=None)  # no read timeout
            self._http_pool = urllib3.PoolManager(retries=urllib3.Retry(total=3), timeout=timeout)
        return self._http_pool

    def gen_transfer_pairs(
        self,
        chunker: Optional[Chunker] = None,
        transfer_config: Optional[TransferConfig] = field(init=False, default_factory=lambda: TransferConfig()),
    ) -> Generator[TransferPair, None, None]:
        """Generate transfer pairs for the transfer job.

        :param chunker: chunker that makes the chunk requests
        :type chunker: Chunker
        """
        if chunker is None:  # used for external access to transfer pair list
            chunker = Chunker(self.src_iface, self.dst_ifaces, transfer_config)  # TODO: should read in existing transfer config
        yield from chunker.transfer_pair_generator(self.src_prefix, self.dst_prefixes, self.recursive, self._pre_filter_fn)

    def dispatch(
        self,
        dataplane: "Dataplane",
        dispatch_batch_size: int = 100,  # 6.4 GB worth of chunks
        transfer_config: Optional[TransferConfig] = field(init=False, default_factory=lambda: TransferConfig()),
    ) -> Generator[Chunk, None, None]:
        """Dispatch transfer job to specified gateways.

        :param dataplane: dataplane that starts the transfer job
        :type dataplane: Dataplane
        :param transfer_config: the configuration during the transfer
        :type transfer_config: TransferConfig
        :param dispatch_batch_size: maximum size of the buffer to temporarily store the generators (default: 1000)
        :type dispatch_batch_size: int
        """
        chunker = Chunker(self.src_iface, self.dst_ifaces, transfer_config)
        transfer_pair_generator = self.gen_transfer_pairs(chunker)  # returns TransferPair objects
        gen_transfer_list = chunker.tail_generator(transfer_pair_generator, self.transfer_list)
        chunks = chunker.chunk(gen_transfer_list)
        batches = chunker.batch_generator(
            chunker.prefetch_generator(chunks, buffer_size=dispatch_batch_size * 32), batch_size=dispatch_batch_size
        )

        # dispatch chunk requests
        src_gateways = dataplane.source_gateways()
        queue_size = [0] * len(src_gateways)
        n_multiparts = 0
        start = time.time()

        for batch in batches:
            # send upload_id mappings to sink gateways
            upload_id_batch = [cr for cr in batch if cr.upload_id_mapping is not None]
            region_dst_gateways = dataplane.sink_gateways()
            for region_tag, dst_gateways in region_dst_gateways.items():
                for dst_gateway in dst_gateways:
                    # collect upload id mappings
                    mappings = {}
                    for message in upload_id_batch:
                        for region_tag, (key, id) in message.upload_id_mapping.items():
                            if region_tag not in mappings:
                                mappings[region_tag] = {}
                            mappings[region_tag][key] = id

                    # send mapping to gateway
                    reply = self.http_pool.request(
                        "POST",
                        f"{dst_gateway.gateway_api_url}/api/v1/upload_id_maps",
                        body=json.dumps(mappings).encode("utf-8"),
                        headers={"Content-Type": "application/json"},
                    )
                    # TODO: assume that only destination nodes would write to the obj store
                    if reply.status != 200:
                        raise Exception(
                            f"Failed to update upload ids to the dst gateway {dst_gateway.instance_name()}: {reply.data.decode('utf-8')}"
                        )

            # send chunk requests to source gateways
            chunk_batch = [cr.chunk for cr in batch if cr.chunk is not None]
            # TODO: allow multiple partition ids per chunk
            for chunk in chunk_batch:  # assign job UUID as partition ID
                chunk.partition_id = self.uuid
            min_idx = queue_size.index(min(queue_size))
            n_added = 0
            while n_added < len(chunk_batch):
                # TODO: should update every source instance queue size
                server = src_gateways[min_idx]
                assert Chunk.from_dict(chunk_batch[0].as_dict()) == chunk_batch[0], f"Invalid chunk request: {chunk_batch[0].as_dict}"

                # TODO: make async
                st = time.time()
                reply = self.http_pool.request(
                    "POST",
                    f"{server.gateway_api_url}/api/v1/chunk_requests",
                    body=json.dumps([chunk.as_dict() for chunk in chunk_batch[n_added:]]).encode("utf-8"),
                    headers={"Content-Type": "application/json"},
                )
                if reply.status != 200:
                    raise Exception(f"Failed to dispatch chunk requests {server.instance_name()}: {reply.data.decode('utf-8')}")
                et = time.time()
                reply_json = json.loads(reply.data.decode("utf-8"))
                n_added += reply_json["n_added"]
                logger.fs.debug(f"Added {n_added} chunks to server {server} in {et-st}: {reply_json}")
                queue_size[min_idx] = reply_json["qsize"]  # update queue size
                # dont try again with some gateway
                min_idx = (min_idx + 1) % len(src_gateways)

            yield from chunk_batch

            # copy new multipart transfers to the multipart transfer list
            updated_len = len(chunker.multipart_upload_requests)
            self.multipart_transfer_list.extend(chunker.multipart_upload_requests[n_multiparts:updated_len])
            n_multiparts = updated_len

    def finalize(self):
        """Complete the multipart upload requests"""
        typer.secho(f"Finalizing multipart uploads...", fg="bright_black")
        groups = defaultdict(list)
        for req in self.multipart_transfer_list:
            if "region" not in req or "bucket" not in req:
                raise Exception(f"Invalid multipart upload request: {req}")
            groups[(req["region"], req["bucket"])].append(req)
        for key, group in groups.items():
            region, bucket = key
            batch_len = max(1, len(group) // 128)
            batches = [group[i : i + batch_len] for i in range(0, len(group), batch_len)]
            obj_store_interface = StorageInterface.create(region, bucket)

            def complete_fn(batch):
                for req in batch:
                    logger.fs.debug(f"Finalize upload id {req['upload_id']} for key {req['key']}")
                    retry_backoff(
                        partial(obj_store_interface.complete_multipart_upload, req["key"], req["upload_id"], req["metadata"]),
                        initial_backoff=0.5,
                    )

            do_parallel(complete_fn, batches, n=8)

        # TODO: Do NOT do this if we are pipelining multiple transfers - remove just what was completed
        self.multipart_transfer_list = []

    def verify(self):
        """Verify the integrity of the transfered destination objects"""

        def verify_region(i):
            dst_iface = self.dst_ifaces[i]
            dst_prefix = self.dst_prefixes[i]

            # gather destination key mapping for this region
            dst_keys = {pair.dst_objs[dst_iface.region_tag()].key: pair.src_obj for pair in self.transfer_list}

            # list and check destination prefix
            for obj in dst_iface.list_objects(dst_prefix):
                # check metadata (src.size == dst.size) && (src.modified <= dst.modified)
                src_obj = dst_keys.get(obj.key)
                if src_obj and src_obj.size == obj.size and src_obj.last_modified <= obj.last_modified:
                    del dst_keys[obj.key]

            if dst_keys:
                # failed_keys = [obj.key for obj in dst_keys.values()]
                failed_keys = list(dst_keys.keys())
                if failed_keys == [""]:
                    return  # ignore empty key
                raise exceptions.TransferFailedException(
                    f"Destination {dst_iface.region_tag()} bucket {dst_iface.bucket()}: {len(dst_keys)} objects failed verification {failed_keys}"
                )

        n = 1  # number threads
        assert n == 1, "Only use one thread for verifying objects: n>1 causes concurrency error"
        do_parallel(
            verify_region,
            range(len(self.dst_ifaces)),
            spinner=True,
            spinner_persist=False,
            desc="Verifying objects in destination buckets",
            n=n,
        )

    def size_gb(self):
        """Return the size of the transfer in GB"""
        total_size = 0
        for pair in self.gen_transfer_pairs():
            total_size += pair.src_obj.size
        return total_size / 1e9


@dataclass
class SyncJob(CopyJob):
    """sync job that copies the source objects that does not exist in the destination bucket to the destination"""

    def __init__(self, src_path: str, dst_paths: List[str] or str, requester_pays: bool = False, job_id: Optional[str] = None):
        super().__init__(src_path, dst_paths, True, requester_pays, job_id)
        self.transfer_list = []
        self.multipart_transfer_list = []

        assert isinstance(self.src_iface, ObjectStoreInterface), "Source must be an object store interface"
        assert not any(
            [not isinstance(iface, ObjectStoreInterface) for iface in self.dst_ifaces]
        ), "Destination must be a object store interface"

    def gen_transfer_pairs(
        self,
        chunker: Optional[Chunker] = None,
        transfer_config: Optional[TransferConfig] = field(init=False, default_factory=lambda: TransferConfig()),
    ) -> Generator[TransferPair, None, None]:
        """Generate transfer pairs for the transfer job.

        :param chunker: chunker that makes the chunk requests
        :type chunker: Chunker
        """
        if chunker is None:  # used for external access to transfer pair list
            chunker = Chunker(self.src_iface, self.dst_ifaces, transfer_config)
        transfer_pair_gen = chunker.transfer_pair_generator(self.src_prefix, self.dst_prefixes, self.recursive, self._pre_filter_fn)

        # only single destination supported
        assert len(self.dst_ifaces) == 1, "Only single destination supported for sync job"

        # enrich destination objects with metadata
        for src_obj, dest_obj in self._enrich_dest_objs(transfer_pair_gen, self.dst_prefixes):
            if self._post_filter_fn(src_obj, dest_obj):
                yield TransferPair(
                    src_obj=src_obj,
                    dst_objs={self.dst_ifaces[0].region_tag(): dest_obj},
                    dst_key=dest_obj.key.replace(self.dst_prefixes[0], ""),
                )

    def _enrich_dest_objs(
        self, transfer_pairs: Generator[TransferPair, None, None], dest_prefixes: List[str]
    ) -> Generator[Tuple[ObjectStoreObject, ObjectStoreObject], None, None]:
        """
        For skyplane sync, we enrich dest obj metadata with our existing dest obj metadata from the dest bucket following a query.

        :meta private:
        :param transfer_pairs: generator of transfer pairs
        :type transfer_pairs: Generator
        """
        for i in range(len(self.dst_ifaces)):
            dst_iface = self.dst_ifaces[i]
            dest_prefix = dest_prefixes[i]
            logger.fs.debug(f"Querying objects in {dst_iface.bucket()}")
            if not hasattr(self, "_found_dest_objs"):
                self._found_dest_objs = {obj.key: obj for obj in dst_iface.list_objects(dest_prefix)}
            for pair in transfer_pairs:
                src_obj = pair.src_obj
                dest_obj = list(pair.dst_objs.values())[0]
                assert len(list(pair.dst_objs.keys())) == 1, f"Multiple destinations are not support for sync: {pair.dst_objs}"
                if dest_obj.key in self._found_dest_objs:
                    dest_obj.size = self._found_dest_objs[dest_obj.key].size
                    dest_obj.last_modified = self._found_dest_objs[dest_obj.key].last_modified
                yield src_obj, dest_obj

    @classmethod
    def _post_filter_fn(cls, src_obj: ObjectStoreObject, dest_obj: ObjectStoreObject) -> bool:
        """Optionally filter destination objects after they are transferred.

        :param src_obj: source object to be transferred
        :type src_obj: ObjectStoreObject
        :param dest_obj: destination object transferred
        :type dest_obj: ObjectStoreObject
        """
        return not dest_obj.exists or (src_obj.last_modified > dest_obj.last_modified or src_obj.size != dest_obj.size)



from typing import Optional, Dict
import json

from skyplane.config import SkyplaneConfig
from skyplane.config_paths import config_path, aws_config_path, aws_quota_path
from skyplane.utils import imports, fn, logger


class AWSAuthentication:
    def __init__(self, config: Optional[SkyplaneConfig] = None, access_key: Optional[str] = None, secret_key: Optional[str] = None):
        """Loads AWS authentication details. If no access key is provided, it will try to load credentials using boto3"""
        if not config is None:
            self.config = config
        else:
            self.config = SkyplaneConfig.load_config(config_path)

        if access_key and secret_key:
            self.config_mode = "manual"
            self._access_key = access_key
            self._secret_key = secret_key
        else:
            self.config_mode = "iam_inferred"
            self._access_key = None
            self._secret_key = None

    def _get_ec2_vm_quota(self, region) -> Dict[str, int]:
        """Given the region, get the maximum number of vCPU that can be launched.

        Returns:
            name_to_quota: a dictionary of quota name to quota value
        """
        # NOTE: the QuotaCode can be retried via
        # aws service-quotas list-service-quotas --service-code ec2
        # The items we are looking is
        # "Running On-Demand Standard (A, C, D, H, I, M, R, T, Z) instances" L-1216C47A
        # "All Standard (A, C, D, H, I, M, R, T, Z) Spot Instance Requests" L-34B43A08
        quotas_client = self.get_boto3_client("service-quotas", region)

        name_to_quota = {}
        for name, code in [("on_demand_standard_vcpus", "L-1216C47A"), ("spot_standard_vcpus", "L-34B43A08")]:
            try:
                retrieved_quota = quotas_client.get_service_quota(ServiceCode="ec2", QuotaCode=code)
                name_to_quota[name] = int(retrieved_quota["Quota"]["Value"])
            except Exception:
                logger.warning(
                    f"Failed to retrieve quota information for {name} in {region}. Skyplane will use a conservative configuration."
                )
        return name_to_quota

    @imports.inject("boto3", pip_extra="aws")
    def save_region_config(boto3, self, config: SkyplaneConfig):
        if not config.aws_enabled:
            self.clear_region_config()
            return
        with aws_config_path.open("w") as f:
            region_list = []

            # query regions
            describe_regions = boto3.client("ec2", region_name="us-east-1").describe_regions()
            for region in describe_regions["Regions"]:
                if region["OptInStatus"] == "opt-in-not-required" or region["OptInStatus"] == "opted-in":
                    region_text = region["Endpoint"]
                    region_name = region_text[region_text.find(".") + 1 : region_text.find(".amazon")]
                    region_list.append(region_name)
            f.write("\n".join(region_list))

        quota_infos = fn.do_parallel(
            self._get_ec2_vm_quota, region_list, return_args=False, spinner=True, desc="Retrieving EC2 Quota information"
        )
        region_infos = [dict(**info, **{"region_name": name}) for name, info in zip(region_list, quota_infos)]
        with aws_quota_path.open("w") as f:
            f.write(json.dumps(region_infos, indent=2))

    def clear_region_config(self):
        with aws_config_path.open("w") as f:
            f.write("")

    @staticmethod
    def get_region_config():
        if not aws_config_path.exists():
            return []
        with aws_config_path.open() as f:
            region_list = []
            for region in f.read().split("\n"):
                region_list.append(region)
            return region_list

    @property
    def access_key(self):
        if self._access_key is None:
            self._access_key, self._secret_key = self.infer_credentials()
        return self._access_key

    @property
    def secret_key(self):
        if self._secret_key is None:
            self._access_key, self._secret_key = self.infer_credentials()
        return self._secret_key

    def enabled(self):
        return self.config.aws_enabled

    @imports.inject("boto3", pip_extra="aws")
    def infer_credentials(boto3, self):
        # todo load temporary credentials from STS
        cached_credential = getattr(self.__cached_credentials, "boto3_credential", None)
        if cached_credential is None:
            session = boto3.Session()
            credentials = session.get_credentials()
            if credentials:
                credentials = credentials.get_frozen_credentials()
                cached_credential = (credentials.access_key, credentials.secret_key)
            setattr(self.__cached_credentials, "boto3_credential", cached_credential)
        return cached_credential if cached_credential else (None, None)

    @imports.inject("boto3", pip_extra="aws")
    def get_boto3_session(boto3, self, aws_region: Optional[str] = None):
        if self.config_mode == "manual":
            return boto3.Session(aws_access_key_id=self.access_key, aws_secret_access_key=self.secret_key, region_name=aws_region)
        else:
            return boto3.Session(region_name=aws_region)

    def get_boto3_resource(self, service_name, aws_region=None):
        return self.get_boto3_session().resource(service_name, region_name=aws_region)

    def get_boto3_client(self, service_name, aws_region=None):
        if aws_region is None:
            return self.get_boto3_session().client(service_name)
        else:
            return self.get_boto3_session().client(service_name, region_name=aws_region)

    def get_azs_in_region(self, region):
        ec2 = self.get_boto3_client("ec2", region)
        azs = []
        for az in ec2.describe_availability_zones()["AvailabilityZones"]:
            azs.append(az["ZoneName"])
        return azs



import base64
import hashlib
import os
from functools import lru_cache

from typing import Any, Iterator, List, Optional, Tuple

from skyplane import exceptions, compute
from skyplane.exceptions import NoSuchObjectException
from skyplane.obj_store.azure_storage_account_interface import AzureStorageAccountInterface
from skyplane.obj_store.object_store_interface import ObjectStoreInterface, ObjectStoreObject
from skyplane.utils import logger, imports
from azure.storage.blob import ContentSettings


MAX_BLOCK_DIGITS = 5


class AzureBlobObject(ObjectStoreObject):
    def full_path(self):
        account_name, container_name = self.bucket.split("/")
        return os.path.join(f"https://{account_name}.blob.core.windows.net", container_name, self.key)


class AzureBlobInterface(ObjectStoreInterface):
    def __init__(self, account_name: str, container_name: str, max_concurrency=1):
        self.auth = compute.AzureAuthentication()
        self.storage_account_interface = AzureStorageAccountInterface(account_name)
        self.account_name = account_name
        self.container_name = container_name
        self.max_concurrency = max_concurrency  # parallel upload/downloads, seems to cause issues if too high

    @property
    def provider(self):
        return "azure"

    def path(self):
        return f"https://{self.account_name}.blob.core.windows.net/{self.container_name}"

    def region_tag(self):
        return f"azure:{self.storage_account_interface.azure_region}"

    def bucket(self) -> str:
        return f"{self.account_name}/{self.container_name}"

    @property
    def blob_service_client(self):
        return self.auth.get_blob_service_client(f"https://{self.account_name}.blob.core.windows.net")

    @property
    def container_client(self):
        return self.auth.get_container_client(f"https://{self.account_name}.blob.core.windows.net", self.container_name)

    @imports.inject("azure.core.exceptions", pip_extra="azure")
    def bucket_exists(exceptions, self):
        try:
            self.container_client.get_container_properties()
            return True
        except exceptions.ResourceNotFoundError:
            return False

    def exists(self, obj_name):
        return self.blob_service_client.get_blob_client(container=self.container_name, blob=obj_name).exists()

    @imports.inject("azure.core.exceptions", pip_extra="azure")
    def create_container(exceptions, self):
        try:
            self.container_client.create_container()
        except exceptions.ResourceExistsError:
            logger.warning(f"Unable to create container {self.container_name} as it already exists")

    def create_bucket(self, azure_region, resource_group=compute.AzureServer.resource_group_name, premium_tier=True):
        tier = "Premium_LRS" if premium_tier else "Standard_LRS"
        if not self.storage_account_interface.storage_account_exists_in_account():
            logger.debug(f"Creating storage account {self.account_name}")
            self.storage_account_interface.create_storage_account(azure_region, resource_group, tier)
        if not self.bucket_exists():
            logger.debug(f"Creating container {self.container_name}")
            self.create_container()

    @imports.inject("azure.core.exceptions", pip_extra="azure")
    def delete_container(exceptions, self):
        try:
            self.container_client.delete_container()
        except exceptions.ResourceNotFoundError:
            logger.warning("Unable to delete container as it doesn't exists")

    def delete_bucket(self):
        return self.delete_container()

    @imports.inject("azure.core.exceptions", pip_extra="azure")
    def list_objects(exceptions, self, prefix="") -> Iterator[AzureBlobObject]:
        blobs = self.container_client.list_blobs(name_starts_with=prefix)
        try:
            for blob in blobs:
                yield AzureBlobObject(
                    blob.name,
                    provider="azure",
                    bucket=f"{self.account_name}/{blob.container}",
                    size=blob.size,
                    last_modified=blob.last_modified,
                    mime_type=getattr(blob.content_settings, "content_type", None),
                )
        except exceptions.HttpResponseError as e:
            if "AuthorizationPermissionMismatch" in str(e):
                logger.error(
                    f"Unable to list objects in container {self.container_name} as you don't have permission to access it. You need the 'Storage Blob Data Contributor' and 'Storage Account Contributor' roles: {e}"
                )
                raise e from None

    def delete_objects(self, keys: List[str]):
        for key in keys:
            blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=key)
            blob_client.delete_blob()

    @lru_cache(maxsize=1024)
    @imports.inject("azure.core.exceptions", pip_extra="azure")
    def get_obj_metadata(exceptions, self, obj_name):
        blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=obj_name)
        try:
            return blob_client.get_blob_properties()
        except exceptions.ResourceNotFoundError as e:
            raise NoSuchObjectException(f"Object {obj_name} does not exist, or you do not have permission to access it") from e

    def get_obj_size(self, obj_name):
        return self.get_obj_metadata(obj_name).size

    def get_obj_last_modified(self, obj_name):
        return self.get_obj_metadata(obj_name).last_modified

    def get_obj_mime_type(self, obj_name):
        return self.get_obj_metadata(obj_name).content_settings.content_type

    def download_object(
        self, src_object_name, dst_file_path, offset_bytes=None, size_bytes=None, write_at_offset=False, generate_md5=False
    ) -> Tuple[Optional[str], Optional[bytes]]:
        src_object_name, dst_file_path = str(src_object_name), str(dst_file_path)
        if size_bytes is not None and offset_bytes is None:
            offset_bytes = 0
        downloader = self.container_client.download_blob(
            src_object_name, offset=offset_bytes, length=size_bytes, max_concurrency=self.max_concurrency
        )

        if not os.path.exists(dst_file_path):
            open(dst_file_path, "a").close()
        if generate_md5:
            m = hashlib.md5()
        with open(dst_file_path, "wb+" if write_at_offset else "wb") as f:
            f.seek(offset_bytes if write_at_offset else 0)
            for b in downloader.chunks():
                if generate_md5:
                    m.update(b)
                f.write(b)
        md5 = m.digest() if generate_md5 else None
        mime_type = self.get_obj_metadata(src_object_name).content_settings.content_type
        return mime_type, md5

    @imports.inject("azure.storage.blob", pip_extra="azure")
    def upload_object(azure_blob, self, src_file_path, dst_object_name, part_number=None, upload_id=None, check_md5=None, mime_type=None):
        """Uses the BlobClient instead of ContainerClient since BlobClient allows for
        block/part level manipulation for multi-part uploads
        """
        src_file_path, dst_object_name = str(src_file_path), str(dst_object_name)
        print(f"Uploading {src_file_path} to {dst_object_name}")

        try:
            blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=dst_object_name)

            # multipart upload
            if part_number is not None and upload_id is not None:
                with open(src_file_path, "rb") as f:
                    block_id = AzureBlobInterface.id_to_base64_encoding(part_number=part_number, dest_key=dst_object_name)
                    blob_client.stage_block(block_id=block_id, data=f, length=os.path.getsize(src_file_path))  # stage the block
                    return

            # single upload
            with open(src_file_path, "rb") as f:
                blob_client.upload_blob(
                    data=f,
                    length=os.path.getsize(src_file_path),
                    max_concurrency=self.max_concurrency,
                    overwrite=True,
                    content_settings=azure_blob.ContentSettings(content_type=mime_type),
                )

                # check MD5 if required
                if check_md5:
                    b64_md5sum = base64.b64encode(check_md5).decode("utf-8") if check_md5 else None
                    blob_md5 = blob_client.get_blob_properties().properties.content_settings.content_md5
                    if b64_md5sum != blob_md5:
                        raise exceptions.ChecksumMismatchException(
                            f"Checksum mismatch for object {dst_object_name} in Azure container {self.container_name}, "
                            + f"expected {b64_md5sum}, got {blob_md5}"
                        )
        except Exception as e:
            raise ValueError(f"Failed to upload {dst_object_name} to bucket {self.container_name} upload id {upload_id}: {e}")

    def initiate_multipart_upload(self, dst_object_name: str, mime_type: Optional[str] = None) -> str:
        """Azure does not have an equivalent function to return an upload ID like s3 and gcs do.
        Blocks in Azure are uploaded and associated with an ID, and can then be committed in a single operation to create the blob.
        We will just return the dst_object_name (blob name) as the "upload_id" to keep the return type consistent for the multipart thread.

        :param dst_object_name: name of the destination object, also our psuedo-uploadID
        :type dst_object_name: str
        :param mime_type: unused in this function but is kept for consistency with the other interfaces (default: None)
        :type mime_type: str
        """

        assert len(dst_object_name) > 0, f"Destination object name must be non-empty: '{dst_object_name}'"

        return dst_object_name

    @imports.inject("azure.storage.blob", pip_extra="azure")
    def complete_multipart_upload(azure_blob, self, dst_object_name: str, upload_id: str, metadata: Optional[Any] = None) -> None:
        """After all blocks of a blob are uploaded/staged with their unique block_id,
        in order to complete the multipart upload, we commit them together.

        :param dst_object_name: name of the destination object, also is used to index into our block mappings
        :type dst_object_name: str
        :param upload_id: upload_id to index into our block id mappings, should be the same as the dst_object_name in Azure
        :type upload_id: str
        :param metadata: In Azure, this custom data is the blockID list (parts) and the object mime_type from the TransferJob instance (default: None)
        :type metadata: Optional[Any]
        """

        assert upload_id == dst_object_name, "In Azure, upload_id should be the same as the blob name."
        assert metadata is not None, "In Azure, the custom data should exist for multipart"

        # Decouple the custom data
        block_list, mime_type = metadata
        assert block_list != [], "The blockID list shouldn't be empty for Azure multipart"
        block_list = list(map(lambda block_id: azure_blob.BlobBlock(block_id=block_id), block_list))

        blob_client = self.blob_service_client.get_blob_client(container=self.container_name, blob=dst_object_name)
        try:
            # The below operation will create the blob from the uploaded blocks.
            blob_client.commit_block_list(block_list=block_list, content_settings=azure_blob.ContentSettings(content_type=mime_type))
        except Exception as e:
            raise exceptions.SkyplaneException(f"Failed to complete multipart upload for {dst_object_name}: {str(e)}")

    @staticmethod
    def id_to_base64_encoding(part_number: int, dest_key: str) -> str:
        """Azure expects all blockIDs to be Base64 strings. This function serves to convert the part numbers to
        base64-encoded strings of the same length. The maximum number of blocks one blob supports in Azure is
        50,000 blocks, so the maximum length to pad zeroes to will be (#digits in 50,000 = len("50000") = 5) + len(dest_key)

        :param part_number: part number of the block, determined while splitting the date into chunks before the transfer
        :type part_number: int
        :param dest_key: destination object key, used to distinguish between different objects during concurrent uploads to the same container
        """
        max_length = MAX_BLOCK_DIGITS + len(dest_key)
        block_id = f"{part_number}{dest_key}"
        block_id = block_id.ljust(max_length, "0")  # pad with zeroes to get consistent length
        block_id = block_id.encode("utf-8")
        block_id = base64.b64encode(block_id).decode("utf-8")
        return block_id
